import logging
from collections.abc import Iterable

from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.script_object import InstanceId, InstanceRef
from retro_data_structures.properties.echoes.archetypes.LayerSwitch import LayerSwitch
from retro_data_structures.properties.echoes.objects import (
    Actor,
    IngSpiderballGuardian,
    MemoryRelay,
    Pickup,
    ScriptLayerController,
    Splinter,
    Timer,
)

from open_prime_rando.echoes.asset_ids import agon_wastes, great_temple, sanctuary_fortress, temple_grounds, torvus_bog
from open_prime_rando.echoes.asset_ids.world import (
    AGON_WASTES_MLVL,
    GREAT_TEMPLE_MLVL,
    SANCTUARY_FORTRESS_MLVL,
    TEMPLE_GROUNDS_MLVL,
    TORVUS_BOG_MLVL,
)
from open_prime_rando.patcher_editor import PatcherEditor

LOG = logging.getLogger("echoes_patcher")


def apply_all(editor: PatcherEditor):
    """
    Applies patches that rebalance aspects of the game for a better rando experience.
    """
    landing_site(editor)
    bionergy_production(editor)
    remove_luminoth_barriers(editor)
    torvus_energy_controller(editor)
    hive_tunnel(editor)
    agon_temple(editor)
    temple_sanctuary(editor)
    main_reactor(editor)
    dynamo_works(editor)
    torvus_temple(editor)
    gfmc_compound(editor)


def landing_site(editor: PatcherEditor):
    """
    Removes the intro cinematic.
    """
    area = editor.get_area(TEMPLE_GROUNDS_MLVL, temple_grounds.LANDING_SITE_MREA)

    remove_intro = True

    # FIXME: use layers API
    area.get_layer("1st Pass - Intro Cinematic").active = not remove_intro
    for layer_name in (
        "Save Station Load",
        "Ship Repair",
        "Luminoth Key Bearer",
        "WAR CHEST"
    ):
        area.get_layer(layer_name).active = remove_intro

    if remove_intro:
        timer = area.get_layer("Default").add_instance_with(Timer(
            time=0.01,
            auto_start=True
        ))
        for inst in (
            "Keep Samus Ship",
            "Savestation Recharge Always Plays",
            "Ambient Music Memory Relay"
        ):
            # TODO: unclear whether the timer is necessary, or if we can just activate these immediately
            timer.add_connection(State.Zero, Message.Activate, area.get_instance(inst))


def bionergy_production(editor: PatcherEditor):
    """
    Deactivate the trigger for flying pirates after killing them all.
    """
    area = editor.get_area(AGON_WASTES_MLVL, agon_wastes.BIOENERGY_PRODUCTION_MREA)

    counter = area.get_instance("Dead Pirates")
    counter.add_connection(State.MaxReached, Message.Deactivate,
                           area.get_instance("Turn On Flying Pirates"))


def _disable_layer_controllers(area: Area, layer_controllers: Iterable[InstanceRef]):
    for inst in layer_controllers:
        with area.get_instance(inst).edit_properties(ScriptLayerController) as controller:
            controller.editor_properties.active = False


def remove_luminoth_barriers(editor: PatcherEditor):
    """
    Remove Luminoth barriers throughout the game.
    """
    # Agon Energy Controller
    agon = editor.get_area(AGON_WASTES_MLVL, agon_wastes.AGON_ENERGY_CONTROLLER_MREA)
    _disable_layer_controllers(agon, ["Increment - 07_Temple - Luminoth Barrier Swamp"])

    # Torvus Energy Controller
    torvus = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.TORVUS_ENERGY_CONTROLLER_MREA)
    _disable_layer_controllers(torvus, ["Increment - 07_Temple - Luminoth Barrier Cliffs"])

    # Dark Agon Energy Controller
    dark_agon = editor.get_area(AGON_WASTES_MLVL, agon_wastes.DARK_AGON_ENERGY_CONTROLLER_MREA)
    _disable_layer_controllers(dark_agon, ["Increment - 0A_Sand_Hall - Luminoth Barriers"])

    # Dark Torvus Energy Controller
    dark_torvus = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.DARK_TORVUS_ENERGY_CONTROLLER_MREA)
    _disable_layer_controllers(dark_torvus, (
        "Increment - 0A_Swamp - Luminoth Barriers",
        "Increment - 04_Swamp - Luminoth Barriers"
    ))

    # Hive Energy Controller
    hive = editor.get_area(SANCTUARY_FORTRESS_MLVL, sanctuary_fortress.HIVE_ENERGY_CONTROLLER_MREA)
    _disable_layer_controllers(hive, (
        "Increment - 0P_Cliff - Luminoth Barriers",
        "Increment - 0A_Cliff - Luminoth Barriers",
        "Increment - 0Q_Cliff - Luminoth Barriers",
    ))


def torvus_energy_controller(editor: PatcherEditor):
    """
    Don't remove the Torvus Temple fight.
    """
    area = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.TORVUS_ENERGY_CONTROLLER_MREA)
    _disable_layer_controllers(area, (
        "DECREMENT 04_Swamp_Temple 1st Pass",
        "Decrement - 04_Swamp_Temple - 1st Pass",
        "Increment - 04_Swamp_Temple - 2ndPass"
    ))


def hive_tunnel(editor: PatcherEditor):
    """
    Unknown purpose.
    """
    area = editor.get_area(TEMPLE_GROUNDS_MLVL, temple_grounds.HIVE_TUNNEL_MREA)

    with area.get_instance("Timer 001").edit_properties(Timer) as timer:
        timer.time = 0.02
    with area.get_instance("Decrement - Webbing").edit_properties(ScriptLayerController) as controller:
        controller.is_dynamic = False


def agon_temple(editor: PatcherEditor):
    """
    No longer locks the doors after the Bomb Guardian fight.
    """
    area = editor.get_area(AGON_WASTES_MLVL, agon_wastes.AGON_TEMPLE_MREA)

    area.get_layer("Lock Doors").active = False # FIXME: use layers API


def temple_sanctuary(editor: PatcherEditor):
    """
    Patches Alpha Splinter to take damage properly from Power Bombs.
    Keep the Emerald gate active from the beginning.
    Fade in the Pickup.
    """
    area = editor.get_area(GREAT_TEMPLE_MLVL, great_temple.TEMPLE_SANCTUARY_MREA)

    with area.get_instance("MEGA Splinter Light").edit_properties(Splinter) as alpha:
        custom_rule = editor._resolve_asset_id("custom_knockback.RULE")
        alpha.patterned.knockback_rules = custom_rule
        alpha.ing_possession_data.ing_vulnerability.power_bomb.damage_multiplier = 3000.0

    with area.get_instance(0x0200B0).edit_properties(MemoryRelay) as relay:
        relay.editor_properties.active = True

    with area.get_instance("Pickup Object").edit_properties(Pickup) as pickup:
        pickup.fadetime = 1.0


def main_reactor(editor: PatcherEditor):
    """
    Change layers properly after DS1's death.
    """
    area = editor.get_area(AGON_WASTES_MLVL, agon_wastes.MAIN_REACTOR_MREA)

    layer_switcher = area.get_instance("Switch Layers To Post-Dark Samus")
    layer_controller = area.get_layer("Default").add_instance_with(ScriptLayerController(
        layer=LayerSwitch(
            area_id=agon_wastes.BIOSTORAGE_STATION_INTERNAL_ID,
            layer_number=3 # 1st Pass
        )
    ))
    layer_switcher.add_connection(State.Zero, Message.Decrement, layer_controller)


def dynamo_works(editor: PatcherEditor):
    """
    Fix Spider Guardian's response to PBs.
    """
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, sanctuary_fortress.DYNAMO_WORKS_MREA)

    with area.get_instance("IngSpiderballGuardian 001").edit_properties(IngSpiderballGuardian) as spider:
        custom_rule = editor._resolve_asset_id("custom_knockback.RULE")
        spider.patterned.knockback_rules = custom_rule


def torvus_temple(editor: PatcherEditor):
    """
    Reduce the size of the barrier such that it doesn't block the path to lower Torvus.
    """
    area = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.TORVUS_TEMPLE_MREA)
    with area.get_instance(0x1B00E0).edit_properties(Actor) as barrier:
        barrier.editor_properties.transform.position.x = -226.4198
        barrier.editor_properties.transform.position.z = 58.7156
        barrier.editor_properties.transform.scale.y = 2.019


def gfmc_compound(editor: PatcherEditor):
    """
    Move the gate to the Default layer.
    """
    area = editor.get_area(TEMPLE_GROUNDS_MLVL, temple_grounds.GFMC_COMPOUND_MREA)

    gate_instances = (
        0x2b00db, 0x2b00dc, 0x2b00ff, 0x2b0101,
        0x2b013c, 0x2b0238, 0x2b0288, 0x2b02e9,
    )
    gate_instances += tuple(range(0x2b0277, 0x2b0287))
    for raw_id in gate_instances:
        # TODO: implement a function to move an instance from one layer to another in RDS
        inst_id = InstanceId(raw_id)
        old_inst = area.get_instance(inst_id)
        new_inst = area.get_layer("Default").add_instance_with(old_inst.get_properties())
        area.remove_instance(old_inst)
        new_inst.id = InstanceId.new(new_inst.id.layer, new_inst.id.area, inst_id.instance)



# Generated by asset_id_files.py

AGON_ENERGY_CONTROLLER_MREA = 0x02FC3717
AGON_MAP_STATION_MREA = 0x845C85A7
AGON_TEMPLE_MREA = 0x75FC9AAE
BATTLEGROUND_MREA = 0xEA79522C
BIOENERGY_PRODUCTION_MREA = 0x13470DA8
BIOSTORAGE_ACCESS_MREA = 0xDEEB1668
BIOSTORAGE_STATION_MREA = 0x80896CE9
BITTER_WELL_MREA = 0x04F1C1D8
CENTRAL_MINING_STATION_MREA = 0xF5A6D972
CENTRAL_STATION_ACCESS_MREA = 0xFE69E0C3
COMMAND_CENTER_MREA = 0x31C25557
COMMAND_CENTER_ACCESS_MREA = 0x708387F9
CONTROLLER_ACCESS_MREA = 0xBF5399C0
CROSSROADS_MREA = 0x8E23DB3F
DARK_AGON_ENERGY_CONTROLLER_MREA = 0xB7DD4006
DARK_AGON_TEMPLE_MREA = 0x7575048A
DARK_AGON_TEMPLE_ACCESS_MREA = 0x2103C8AE
DARK_CONTROLLER_ACCESS_MREA = 0xA2F347D1
DARK_OASIS_MREA = 0x99FB957F
DARK_TRANSIT_STATION_MREA = 0xD73ECD68
DOOMED_ENTRY_MREA = 0xA49F8A6D
DOUBLE_PATH_MREA = 0x7FEF433B
DUELLING_RANGE_MREA = 0x3F5CDF39
FEEDING_PIT_MREA = 0x075A76EB
FEEDING_PIT_ACCESS_MREA = 0xE035C0A5
HALL_OF_STAIRS_MREA = 0x6D164780
ING_CACHE_1_MREA = 0x52962FB6
ING_CACHE_2_MREA = 0x9318F076
ING_CACHE_3_MREA = 0xBA23E5F5
ING_CACHE_4_MREA = 0xC5C7C9D3
JUDGMENT_PIT_MREA = 0x6B77F9BD
JUNCTION_SITE_MREA = 0x1608C563
MAIN_ENERGY_CONTROLLER_MREA = 0xD92890C3
MAIN_REACTOR_MREA = 0xCCD9F39E
MINE_SHAFT_MREA = 0xD7A15431
MINING_PLAZA_MREA = 0x427FAD9A
MINING_STATION_A_MREA = 0xE3BEF27F
MINING_STATION_ACCESS_MREA = 0x8D7A7F8C
MINING_STATION_B_MREA = 0xDB7B2CED
OASIS_ACCESS_MREA = 0xF2CCC41E
PHAZON_SITE_MREA = 0x200EF5E7
PLAZA_ACCESS_MREA = 0xADF88927
PORTAL_ACCESS_MREA = 0x89D246FD
PORTAL_ACCESS_A_MREA = 0x2312EE1D
PORTAL_SITE_MREA = 0xBF7F572C
PORTAL_TERMINAL_MREA = 0x2BCD44A7
SAND_CACHE_MREA = 0xF4B1F727
SAND_PROCESSING_MREA = 0xA4B2CB7E
SANDCANYON_MREA = 0xE5B72628
SAVE_STATION_1_MREA = 0x5A154E2E
SAVE_STATION_2_MREA = 0xAC204F8B
SAVE_STATION_3_MREA = 0x0279F7EF
SAVE_STATION_A_MREA = 0x50017152
SAVE_STATION_C_MREA = 0xB979562C
SECURITY_STATION_A_MREA = 0xF723A29A
SECURITY_STATION_B_MREA = 0x0ADA5AEF
STORAGE_A_MREA = 0x10323944
STORAGE_B_MREA = 0x96A64BEA
STORAGE_C_MREA = 0x5DFA984F
STORAGE_D_MREA = 0x40FFA8F7
TEMPLE_ACCESS_MREA = 0x2A58AC44
TRANSIT_STATION_MREA = 0x79C9C5A0
TRANSPORT_CENTER_MREA = 0x4BDFB7B9
TRANSPORT_TO_SANCTUARY_FORTRESS_MREA = 0xC68B5B51
TRANSPORT_TO_TEMPLE_GROUNDS_MREA = 0x57CE3A52
TRANSPORT_TO_TORVUS_BOG_MREA = 0xA74EC002
TRIAL_GROUNDS_MREA = 0x86C1AB37
TRIAL_TUNNEL_MREA = 0x48E44EF6
VENTILATION_AREA_A_MREA = 0x84303DD5
VENTILATION_AREA_B_MREA = 0x594B330B
WARRIORS_WALK_MREA = 0x33FACC15
WATERING_HOLE_MREA = 0x9B2B4246

NAME_TO_ID_MREA = {
    "Agon Energy Controller": 0x02FC3717,
    "Agon Map Station": 0x845C85A7,
    "Agon Temple": 0x75FC9AAE,
    "Battleground": 0xEA79522C,
    "Bioenergy Production": 0x13470DA8,
    "Biostorage Access": 0xDEEB1668,
    "Biostorage Station": 0x80896CE9,
    "Bitter Well": 0x04F1C1D8,
    "Central Mining Station": 0xF5A6D972,
    "Central Station Access": 0xFE69E0C3,
    "Command Center": 0x31C25557,
    "Command Center Access": 0x708387F9,
    "Controller Access": 0xBF5399C0,
    "Crossroads": 0x8E23DB3F,
    "Dark Agon Energy Controller": 0xB7DD4006,
    "Dark Agon Temple": 0x7575048A,
    "Dark Agon Temple Access": 0x2103C8AE,
    "Dark Controller Access": 0xA2F347D1,
    "Dark Oasis": 0x99FB957F,
    "Dark Transit Station": 0xD73ECD68,
    "Doomed Entry": 0xA49F8A6D,
    "Double Path": 0x7FEF433B,
    "Duelling Range": 0x3F5CDF39,
    "Feeding Pit": 0x075A76EB,
    "Feeding Pit Access": 0xE035C0A5,
    "Hall of Stairs": 0x6D164780,
    "Ing Cache 1": 0x52962FB6,
    "Ing Cache 2": 0x9318F076,
    "Ing Cache 3": 0xBA23E5F5,
    "Ing Cache 4": 0xC5C7C9D3,
    "Judgment Pit": 0x6B77F9BD,
    "Junction Site": 0x1608C563,
    "Main Energy Controller": 0xD92890C3,
    "Main Reactor": 0xCCD9F39E,
    "Mine Shaft": 0xD7A15431,
    "Mining Plaza": 0x427FAD9A,
    "Mining Station A": 0xE3BEF27F,
    "Mining Station Access": 0x8D7A7F8C,
    "Mining Station B": 0xDB7B2CED,
    "Oasis Access": 0xF2CCC41E,
    "Phazon Site": 0x200EF5E7,
    "Plaza Access": 0xADF88927,
    "Portal Access": 0x89D246FD,
    "Portal Access A": 0x2312EE1D,
    "Portal Site": 0xBF7F572C,
    "Portal Terminal": 0x2BCD44A7,
    "Sand Cache": 0xF4B1F727,
    "Sand Processing": 0xA4B2CB7E,
    "Sandcanyon": 0xE5B72628,
    "Save Station 1": 0x5A154E2E,
    "Save Station 2": 0xAC204F8B,
    "Save Station 3": 0x0279F7EF,
    "Save Station A": 0x50017152,
    "Save Station C": 0xB979562C,
    "Security Station A": 0xF723A29A,
    "Security Station B": 0x0ADA5AEF,
    "Storage A": 0x10323944,
    "Storage B": 0x96A64BEA,
    "Storage C": 0x5DFA984F,
    "Storage D": 0x40FFA8F7,
    "Temple Access": 0x2A58AC44,
    "Transit Station": 0x79C9C5A0,
    "Transport Center": 0x4BDFB7B9,
    "Transport to Sanctuary Fortress": 0xC68B5B51,
    "Transport to Temple Grounds": 0x57CE3A52,
    "Transport to Torvus Bog": 0xA74EC002,
    "Trial Grounds": 0x86C1AB37,
    "Trial Tunnel": 0x48E44EF6,
    "Ventilation Area A": 0x84303DD5,
    "Ventilation Area B": 0x594B330B,
    "Warrior's Walk": 0x33FACC15,
    "Watering Hole": 0x9B2B4246,
}
# Generated by asset_id_files.py

AGON_ENERGY_CONTROLLER_MAPA = 0x2B29CC2A
AGON_MAP_STATION_MAPA = 0xAD897E9A
AGON_TEMPLE_MAPA = 0x5C296193
BATTLEGROUND_MAPA = 0xC3ACA911
BIOENERGY_PRODUCTION_MAPA = 0x3A92F695
BIOSTORAGE_ACCESS_MAPA = 0xF73EED55
BIOSTORAGE_STATION_MAPA = 0xA95C97D4
BITTER_WELL_MAPA = 0x2D243AE5
CENTRAL_MINING_STATION_MAPA = 0xDC73224F
CENTRAL_STATION_ACCESS_MAPA = 0xD7BC1BFE
COMMAND_CENTER_MAPA = 0x1817AE6A
COMMAND_CENTER_ACCESS_MAPA = 0x59567CC4
CONTROLLER_ACCESS_MAPA = 0x968662FD
CROSSROADS_MAPA = 0xA7F62002
DARK_AGON_ENERGY_CONTROLLER_MAPA = 0x9E08BB3B
DARK_AGON_TEMPLE_MAPA = 0x5CA0FFB7
DARK_AGON_TEMPLE_ACCESS_MAPA = 0x08D63393
DARK_CONTROLLER_ACCESS_MAPA = 0x8B26BCEC
DARK_OASIS_MAPA = 0xB02E6E42
DARK_TRANSIT_STATION_MAPA = 0xFEEB3655
DOOMED_ENTRY_MAPA = 0x8D4A7150
DOUBLE_PATH_MAPA = 0x563AB806
DUELLING_RANGE_MAPA = 0x16892404
FEEDING_PIT_MAPA = 0x2E8F8DD6
FEEDING_PIT_ACCESS_MAPA = 0xC9E03B98
HALL_OF_STAIRS_MAPA = 0x44C3BCBD
ING_CACHE_1_MAPA = 0x7B43D48B
ING_CACHE_2_MAPA = 0xBACD0B4B
ING_CACHE_3_MAPA = 0x93F61EC8
ING_CACHE_4_MAPA = 0xEC1232EE
JUDGMENT_PIT_MAPA = 0x42A20280
JUNCTION_SITE_MAPA = 0x3FDD3E5E
MAIN_ENERGY_CONTROLLER_MAPA = 0xF0FD6BFE
MAIN_REACTOR_MAPA = 0xE50C08A3
MINE_SHAFT_MAPA = 0xFE74AF0C
MINING_PLAZA_MAPA = 0x6BAA56A7
MINING_STATION_A_MAPA = 0xCA6B0942
MINING_STATION_ACCESS_MAPA = 0xA4AF84B1
MINING_STATION_B_MAPA = 0xF2AED7D0
OASIS_ACCESS_MAPA = 0xDB193F23
PHAZON_SITE_MAPA = 0x09DB0EDA
PLAZA_ACCESS_MAPA = 0x842D721A
PORTAL_ACCESS_MAPA = 0xA007BDC0
PORTAL_ACCESS_A_MAPA = 0x0AC71520
PORTAL_SITE_MAPA = 0x96AAAC11
PORTAL_TERMINAL_MAPA = 0x0218BF9A
SAND_CACHE_MAPA = 0xDD640C1A
SAND_PROCESSING_MAPA = 0x8D673043
SANDCANYON_MAPA = 0xCC62DD15
SAVE_STATION_1_MAPA = 0x73C0B513
SAVE_STATION_2_MAPA = 0x85F5B4B6
SAVE_STATION_3_MAPA = 0x2BAC0CD2
SAVE_STATION_A_MAPA = 0x79D48A6F
SAVE_STATION_C_MAPA = 0x90ACAD11
SECURITY_STATION_A_MAPA = 0xDEF659A7
SECURITY_STATION_B_MAPA = 0x230FA1D2
STORAGE_A_MAPA = 0x39E7C279
STORAGE_B_MAPA = 0xBF73B0D7
STORAGE_C_MAPA = 0x742F6372
STORAGE_D_MAPA = 0x692A53CA
TEMPLE_ACCESS_MAPA = 0x038D5779
TRANSIT_STATION_MAPA = 0x501C3E9D
TRANSPORT_CENTER_MAPA = 0x620A4C84
TRANSPORT_TO_SANCTUARY_FORTRESS_MAPA = 0xEF5EA06C
TRANSPORT_TO_TEMPLE_GROUNDS_MAPA = 0x7E1BC16F
TRANSPORT_TO_TORVUS_BOG_MAPA = 0x8E9B3B3F
TRIAL_GROUNDS_MAPA = 0xAF14500A
TRIAL_TUNNEL_MAPA = 0x6131B5CB
VENTILATION_AREA_A_MAPA = 0xADE5C6E8
VENTILATION_AREA_B_MAPA = 0x709EC836
WARRIORS_WALK_MAPA = 0x1A2F3728
WATERING_HOLE_MAPA = 0xB2FEB97B

NAME_TO_ID_MAPA = {
    "Agon Energy Controller": 0x2B29CC2A,
    "Agon Map Station": 0xAD897E9A,
    "Agon Temple": 0x5C296193,
    "Battleground": 0xC3ACA911,
    "Bioenergy Production": 0x3A92F695,
    "Biostorage Access": 0xF73EED55,
    "Biostorage Station": 0xA95C97D4,
    "Bitter Well": 0x2D243AE5,
    "Central Mining Station": 0xDC73224F,
    "Central Station Access": 0xD7BC1BFE,
    "Command Center": 0x1817AE6A,
    "Command Center Access": 0x59567CC4,
    "Controller Access": 0x968662FD,
    "Crossroads": 0xA7F62002,
    "Dark Agon Energy Controller": 0x9E08BB3B,
    "Dark Agon Temple": 0x5CA0FFB7,
    "Dark Agon Temple Access": 0x08D63393,
    "Dark Controller Access": 0x8B26BCEC,
    "Dark Oasis": 0xB02E6E42,
    "Dark Transit Station": 0xFEEB3655,
    "Doomed Entry": 0x8D4A7150,
    "Double Path": 0x563AB806,
    "Duelling Range": 0x16892404,
    "Feeding Pit": 0x2E8F8DD6,
    "Feeding Pit Access": 0xC9E03B98,
    "Hall of Stairs": 0x44C3BCBD,
    "Ing Cache 1": 0x7B43D48B,
    "Ing Cache 2": 0xBACD0B4B,
    "Ing Cache 3": 0x93F61EC8,
    "Ing Cache 4": 0xEC1232EE,
    "Judgment Pit": 0x42A20280,
    "Junction Site": 0x3FDD3E5E,
    "Main Energy Controller": 0xF0FD6BFE,
    "Main Reactor": 0xE50C08A3,
    "Mine Shaft": 0xFE74AF0C,
    "Mining Plaza": 0x6BAA56A7,
    "Mining Station A": 0xCA6B0942,
    "Mining Station Access": 0xA4AF84B1,
    "Mining Station B": 0xF2AED7D0,
    "Oasis Access": 0xDB193F23,
    "Phazon Site": 0x09DB0EDA,
    "Plaza Access": 0x842D721A,
    "Portal Access": 0xA007BDC0,
    "Portal Access A": 0x0AC71520,
    "Portal Site": 0x96AAAC11,
    "Portal Terminal": 0x0218BF9A,
    "Sand Cache": 0xDD640C1A,
    "Sand Processing": 0x8D673043,
    "Sandcanyon": 0xCC62DD15,
    "Save Station 1": 0x73C0B513,
    "Save Station 2": 0x85F5B4B6,
    "Save Station 3": 0x2BAC0CD2,
    "Save Station A": 0x79D48A6F,
    "Save Station C": 0x90ACAD11,
    "Security Station A": 0xDEF659A7,
    "Security Station B": 0x230FA1D2,
    "Storage A": 0x39E7C279,
    "Storage B": 0xBF73B0D7,
    "Storage C": 0x742F6372,
    "Storage D": 0x692A53CA,
    "Temple Access": 0x038D5779,
    "Transit Station": 0x501C3E9D,
    "Transport Center": 0x620A4C84,
    "Transport to Sanctuary Fortress": 0xEF5EA06C,
    "Transport to Temple Grounds": 0x7E1BC16F,
    "Transport to Torvus Bog": 0x8E9B3B3F,
    "Trial Grounds": 0xAF14500A,
    "Trial Tunnel": 0x6131B5CB,
    "Ventilation Area A": 0xADE5C6E8,
    "Ventilation Area B": 0x709EC836,
    "Warrior's Walk": 0x1A2F3728,
    "Watering Hole": 0xB2FEB97B,
}
# Generated by asset_id_files.py

AGON_ENERGY_CONTROLLER_INTERNAL_ID = 0xE250F791
AGON_MAP_STATION_INTERNAL_ID = 0x538BF04F
AGON_TEMPLE_INTERNAL_ID = 0xE13DA78B
BATTLEGROUND_INTERNAL_ID = 0xC307E36F
BIOENERGY_PRODUCTION_INTERNAL_ID = 0x7D231C5D
BIOSTORAGE_ACCESS_INTERNAL_ID = 0x4ECF9934
BIOSTORAGE_STATION_INTERNAL_ID = 0x7C8E5C4D
BITTER_WELL_INTERNAL_ID = 0xAC9DB4AE
CENTRAL_MINING_STATION_INTERNAL_ID = 0xD16D7648
CENTRAL_STATION_ACCESS_INTERNAL_ID = 0xABAF67E7
COMMAND_CENTER_INTERNAL_ID = 0xAA657163
COMMAND_CENTER_ACCESS_INTERNAL_ID = 0xF681BFC1
CONTROLLER_ACCESS_INTERNAL_ID = 0x4B685014
CROSSROADS_INTERNAL_ID = 0xD9FD4BCB
DARK_AGON_ENERGY_CONTROLLER_INTERNAL_ID = 0xEF855B84
DARK_AGON_TEMPLE_INTERNAL_ID = 0x8DE34ED5
DARK_AGON_TEMPLE_ACCESS_INTERNAL_ID = 0x8627988A
DARK_CONTROLLER_ACCESS_INTERNAL_ID = 0x350DAA91
DARK_OASIS_INTERNAL_ID = 0xD5B16346
DARK_TRANSIT_STATION_INTERNAL_ID = 0x99D40591
DOOMED_ENTRY_INTERNAL_ID = 0x9244D4D5
DOUBLE_PATH_INTERNAL_ID = 0x4EDCF903
DUELLING_RANGE_INTERNAL_ID = 0x1BA88A01
FEEDING_PIT_INTERNAL_ID = 0x2B9333CF
FEEDING_PIT_ACCESS_INTERNAL_ID = 0xA6333FA7
HALL_OF_STAIRS_INTERNAL_ID = 0x5B81EF11
ING_CACHE_1_INTERNAL_ID = 0x6845CF3A
ING_CACHE_2_INTERNAL_ID = 0x765DA1AE
ING_CACHE_3_INTERNAL_ID = 0x2C14A842
ING_CACHE_4_INTERNAL_ID = 0x8C891383
JUDGMENT_PIT_INTERNAL_ID = 0x2BB2B2D4
JUNCTION_SITE_INTERNAL_ID = 0xB9C0A2BC
MAIN_ENERGY_CONTROLLER_INTERNAL_ID = 0x9C550062
MAIN_REACTOR_INTERNAL_ID = 0x80CBA3E1
MINE_SHAFT_INTERNAL_ID = 0xB80C0AB9
MINING_PLAZA_INTERNAL_ID = 0xC6102F1B
MINING_STATION_A_INTERNAL_ID = 0xC3FE1203
MINING_STATION_ACCESS_INTERNAL_ID = 0x69894388
MINING_STATION_B_INTERNAL_ID = 0x3A9B7C8D
OASIS_ACCESS_INTERNAL_ID = 0xB36E29B5
PHAZON_SITE_INTERNAL_ID = 0x3AC464B1
PLAZA_ACCESS_INTERNAL_ID = 0x8CE9BD5B
PORTAL_ACCESS_INTERNAL_ID = 0x512F6418
PORTAL_ACCESS_A_INTERNAL_ID = 0xD1C7657D
PORTAL_SITE_INTERNAL_ID = 0x3EFE8B8D
PORTAL_TERMINAL_INTERNAL_ID = 0x9ECAA95E
SAND_CACHE_INTERNAL_ID = 0x6E1F159C
SAND_PROCESSING_INTERNAL_ID = 0x7A2A2ED6
SANDCANYON_INTERNAL_ID = 0x6B9AD531
SAVE_STATION_1_INTERNAL_ID = 0x951030E9
SAVE_STATION_2_INTERNAL_ID = 0x7B95483C
SAVE_STATION_3_INTERNAL_ID = 0xE92426C4
SAVE_STATION_A_INTERNAL_ID = 0x13E14112
SAVE_STATION_C_INTERNAL_ID = 0xDE113070
SECURITY_STATION_A_INTERNAL_ID = 0x5D6CF46A
SECURITY_STATION_B_INTERNAL_ID = 0xC2640823
STORAGE_A_INTERNAL_ID = 0x484C1B94
STORAGE_B_INTERNAL_ID = 0x3694CC04
STORAGE_C_INTERNAL_ID = 0x1CDC8174
STORAGE_D_INTERNAL_ID = 0xCB256324
TEMPLE_ACCESS_INTERNAL_ID = 0x2704F6F0
TRANSIT_STATION_INTERNAL_ID = 0x00422C4C
TRANSPORT_CENTER_INTERNAL_ID = 0xD3D4F3C4
TRANSPORT_TO_SANCTUARY_FORTRESS_INTERNAL_ID = 0xE6B9F017
TRANSPORT_TO_TEMPLE_GROUNDS_INTERNAL_ID = 0x0B3919B6
TRANSPORT_TO_TORVUS_BOG_INTERNAL_ID = 0x15B1250D
TRIAL_GROUNDS_INTERNAL_ID = 0x47B6E9D0
TRIAL_TUNNEL_INTERNAL_ID = 0x713BC335
VENTILATION_AREA_A_INTERNAL_ID = 0x9F4AD005
VENTILATION_AREA_B_INTERNAL_ID = 0xE522D29F
WARRIORS_WALK_INTERNAL_ID = 0x937A8E98
WATERING_HOLE_INTERNAL_ID = 0x4472720A

NAME_TO_ID_INTERNAL_ID = {
    "Agon Energy Controller": 0xE250F791,
    "Agon Map Station": 0x538BF04F,
    "Agon Temple": 0xE13DA78B,
    "Battleground": 0xC307E36F,
    "Bioenergy Production": 0x7D231C5D,
    "Biostorage Access": 0x4ECF9934,
    "Biostorage Station": 0x7C8E5C4D,
    "Bitter Well": 0xAC9DB4AE,
    "Central Mining Station": 0xD16D7648,
    "Central Station Access": 0xABAF67E7,
    "Command Center": 0xAA657163,
    "Command Center Access": 0xF681BFC1,
    "Controller Access": 0x4B685014,
    "Crossroads": 0xD9FD4BCB,
    "Dark Agon Energy Controller": 0xEF855B84,
    "Dark Agon Temple": 0x8DE34ED5,
    "Dark Agon Temple Access": 0x8627988A,
    "Dark Controller Access": 0x350DAA91,
    "Dark Oasis": 0xD5B16346,
    "Dark Transit Station": 0x99D40591,
    "Doomed Entry": 0x9244D4D5,
    "Double Path": 0x4EDCF903,
    "Duelling Range": 0x1BA88A01,
    "Feeding Pit": 0x2B9333CF,
    "Feeding Pit Access": 0xA6333FA7,
    "Hall of Stairs": 0x5B81EF11,
    "Ing Cache 1": 0x6845CF3A,
    "Ing Cache 2": 0x765DA1AE,
    "Ing Cache 3": 0x2C14A842,
    "Ing Cache 4": 0x8C891383,
    "Judgment Pit": 0x2BB2B2D4,
    "Junction Site": 0xB9C0A2BC,
    "Main Energy Controller": 0x9C550062,
    "Main Reactor": 0x80CBA3E1,
    "Mine Shaft": 0xB80C0AB9,
    "Mining Plaza": 0xC6102F1B,
    "Mining Station A": 0xC3FE1203,
    "Mining Station Access": 0x69894388,
    "Mining Station B": 0x3A9B7C8D,
    "Oasis Access": 0xB36E29B5,
    "Phazon Site": 0x3AC464B1,
    "Plaza Access": 0x8CE9BD5B,
    "Portal Access": 0x512F6418,
    "Portal Access A": 0xD1C7657D,
    "Portal Site": 0x3EFE8B8D,
    "Portal Terminal": 0x9ECAA95E,
    "Sand Cache": 0x6E1F159C,
    "Sand Processing": 0x7A2A2ED6,
    "Sandcanyon": 0x6B9AD531,
    "Save Station 1": 0x951030E9,
    "Save Station 2": 0x7B95483C,
    "Save Station 3": 0xE92426C4,
    "Save Station A": 0x13E14112,
    "Save Station C": 0xDE113070,
    "Security Station A": 0x5D6CF46A,
    "Security Station B": 0xC2640823,
    "Storage A": 0x484C1B94,
    "Storage B": 0x3694CC04,
    "Storage C": 0x1CDC8174,
    "Storage D": 0xCB256324,
    "Temple Access": 0x2704F6F0,
    "Transit Station": 0x00422C4C,
    "Transport Center": 0xD3D4F3C4,
    "Transport to Sanctuary Fortress": 0xE6B9F017,
    "Transport to Temple Grounds": 0x0B3919B6,
    "Transport to Torvus Bog": 0x15B1250D,
    "Trial Grounds": 0x47B6E9D0,
    "Trial Tunnel": 0x713BC335,
    "Ventilation Area A": 0x9F4AD005,
    "Ventilation Area B": 0xE522D29F,
    "Warrior's Walk": 0x937A8E98,
    "Watering Hole": 0x4472720A,
}

DOCK_NAMES = {
    "Agon Energy Controller": {
        "South": 0,
        "North1 (to 08 Copy)": 1,
    },
    "Agon Map Station": {
        "South": 0,
    },
    "Agon Temple": {
        "West": 0,
        "North": 1,
        "East": 2,
        "South": 3,
    },
    "Battleground": {
        "East_Lower": 0,
        "East_Upper": 1,
        "West": 2,
    },
    "Bioenergy Production": {
        "South": 0,
        "North": 1,
        "East": 2,
    },
    "Biostorage Access": {
        "South": 0,
        "North": 1,
    },
    "Biostorage Station": {
        "East": 0,
        "North": 1,
        "West": 2,
    },
    "Bitter Well": {
        "South": 0,
        "North": 1,
    },
    "Central Mining Station": {
        "East_Lower": 0,
        "East_Upper": 1,
        "West": 2,
    },
    "Central Station Access": {
        "West": 0,
        "East": 1,
    },
    "Command Center": {
        "WestLowerDoor": 0,
        "South": 1,
        "North": 2,
        "NorthObservation": 3,
        "WestUpperDoor": 4,
        "WestMorphballTunnel": 5,
        "VirtualNorth": 6,
    },
    "Command Center Access": {
        "EastLowerDoor": 0,
        "EastUpperDoor": 1,
        "EastMorphballTunnel": 2,
        "West_Upper": 3,
        "West_Lower": 4,
    },
    "Controller Access": {
        "North": 0,
        "South1": 1,
    },
    "Crossroads": {
        "West": 0,
        "East": 1,
        "South": 2,
        "VirtualNorth": 3,
    },
    "Dark Agon Energy Controller": {
        "South": 0,
        "North1": 1,
    },
    "Dark Agon Temple": {
        "West": 0,
        "North": 1,
        "East": 2,
        "South": 3,
    },
    "Dark Agon Temple Access": {
        "North": 0,
        "South": 1,
    },
    "Dark Controller Access": {
        "North": 0,
        "South1": 1,
    },
    "Dark Oasis": {
        "East_TopSouth": 0,
        "West_TopSouth": 1,
        "West_TopNorth": 2,
        "South1": 3,
        "VirtualEast": 4,
    },
    "Dark Transit Station": {
        "East": 0,
        "South": 1,
    },
    "Doomed Entry": {
        "WestLowerDoor": 0,
        "South": 1,
        "NorthObservation": 2,
        "WestUpperDoor": 3,
        "VirtualNorth": 4,
    },
    "Double Path": {
        "EastLowerDoor": 0,
        "EastUpperDoor": 1,
        "EastMorphballTunnel": 2,
        "West_Upper": 3,
        "West_Lower": 4,
    },
    "Duelling Range": {
        "North": 0,
        "South": 1,
        "West": 2,
        "East": 3,
        "VirtualDock": 4,
    },
    "Feeding Pit": {
        "East": 0,
        "North": 1,
        "West": 2,
    },
    "Feeding Pit Access": {
        "South": 0,
        "North": 1,
    },
    "Hall of Stairs": {
        "East": 0,
        "West": 1,
    },
    "Ing Cache 1": {
        "East": 0,
    },
    "Ing Cache 2": {
        "West": 0,
    },
    "Ing Cache 3": {
        "East": 0,
    },
    "Ing Cache 4": {
        "West": 0,
        "East": 1,
    },
    "Judgment Pit": {
        "North": 0,
        "East1": 1,
        "South_save": 2,
        "West1": 3,
        "South1": 4,
    },
    "Junction Site": {
        "East": 0,
        "West": 1,
    },
    "Main Energy Controller": {
        "SouthLight1": 0,
        "VirtualNorth": 1,
        "North1 (to 00 Sand Generator)": 2,
    },
    "Main Reactor": {
        "East_TopSouth": 0,
        "West_TopSouth": 1,
        "West_TopNorth": 2,
        "Docks_Docks_Docks_South": 3,
        "VirtualEast": 4,
    },
    "Mine Shaft": {
        "West": 0,
        "East": 1,
    },
    "Mining Plaza": {
        "NorthMap": 0,
        "North": 1,
        "South": 2,
        "West": 3,
        "East1": 4,
        "VirtualDock": 5,
    },
    "Mining Station A": {
        "North": 0,
        "East1": 1,
        "South_save": 2,
        "West1": 3,
        "South1": 4,
    },
    "Mining Station Access": {
        "East": 0,
        "West": 1,
    },
    "Mining Station B": {
        "Docks_SouthLower": 0,
        "East": 1,
        "West": 2,
        "VirtualNorth": 3,
    },
    "Oasis Access": {
        "South": 0,
        "South_Observation": 1,
        "North": 2,
    },
    "Phazon Site": {
        "South": 0,
        "North": 1,
        "East": 2,
    },
    "Plaza Access": {
        "West": 0,
        "East": 1,
    },
    "Portal Access": {
        "South": 0,
        "North": 1,
    },
    "Portal Access A": {
        "South": 0,
        "North": 1,
    },
    "Portal Site": {
        "North": 0,
        "West": 1,
        "VirtualNorth1": 2,
    },
    "Portal Terminal": {
        "North": 0,
        "West": 1,
        "VirtualNorth1": 2,
    },
    "Sand Cache": {
        "North": 0,
    },
    "Sand Processing": {
        "East": 0,
        "West": 1,
    },
    "Sandcanyon": {
        "West": 0,
        "East": 1,
    },
    "Save Station 1": {
        "North": 0,
    },
    "Save Station 2": {
        "North": 0,
        "East": 1,
    },
    "Save Station 3": {
        "South": 0,
        "West": 1,
    },
    "Save Station A": {
        "North": 0,
        "East": 1,
    },
    "Save Station C": {
        "South": 0,
        "West": 1,
    },
    "Security Station A": {
        "West": 0,
        "North": 1,
    },
    "Security Station B": {
        "South": 0,
        "South_Observation": 1,
        "North": 2,
    },
    "Storage A": {
        "North": 0,
    },
    "Storage B": {
        "East": 0,
    },
    "Storage C": {
        "West": 0,
    },
    "Storage D": {
        "East": 0,
    },
    "Temple Access": {
        "North": 0,
        "South": 1,
    },
    "Transit Station": {
        "East": 0,
        "South": 1,
    },
    "Transport Center": {
        "West": 0,
        "East": 1,
        "South": 2,
        "VirtualNorth": 3,
    },
    "Transport to Sanctuary Fortress": {
        "Dock - South": 0,
    },
    "Transport to Temple Grounds": {
        "East": 0,
    },
    "Transport to Torvus Bog": {
        "Dock - North": 0,
    },
    "Trial Grounds": {
        "Docks_SouthLower": 0,
        "East": 1,
        "West": 2,
        "VirtualNorth": 3,
    },
    "Trial Tunnel": {
        "West": 0,
        "East": 1,
    },
    "Ventilation Area A": {
        "docks_North": 0,
        "docks_docks_East": 1,
        "docks_West": 2,
    },
    "Ventilation Area B": {
        "South": 0,
        "North": 1,
    },
    "Warrior's Walk": {
        "West": 0,
        "East": 1,
    },
    "Watering Hole": {
        "West": 0,
        "North": 1,
    },
}




from construct import Container
from retro_data_structures.base_resource import AssetId, RawResource
from retro_data_structures.formats.cmdl import Cmdl
from retro_data_structures.game_check import Game

from open_prime_rando.echoes.custom_assets import custom_asset_path
from open_prime_rando.echoes.dock_lock_rando import dock_type
from open_prime_rando.echoes.dock_lock_rando.dock_type_database import DOCK_TYPES
from open_prime_rando.patcher_editor import PatcherEditor


def add_custom_models(editor: PatcherEditor):
    assets = custom_asset_path().joinpath("doors")
    def get_txtr(n: str, must_exist: bool = True) -> AssetId:
        f = assets.joinpath(n)
        if not must_exist and not f.exists():
            return None
        res = RawResource(
            type="TXTR",
            data=f.read_bytes()
        )
        return editor.add_new_asset(n, res)

    greyscale_emissive = get_txtr("custom_door_lock_greyscale_emissive.TXTR")
    template = editor.get_parsed_asset(0xF115F575, type_hint=Cmdl)

    for door_type in DOCK_TYPES.values():
        if not (
            isinstance(door_type, dock_type.BlastShieldDoorType)
            and isinstance(door_type.shield_model, str)
        ):
            continue
        name = door_type.shield_model
        txtr = get_txtr(f"custom_door_lock_{name}.TXTR")
        emissive = get_txtr(f"custom_door_lock_{name}_emissive.TXTR", must_exist=False)
        if emissive is None:
            emissive = greyscale_emissive

        cmdl = Container(template.raw)
        cmdl.material_sets[0].texture_file_ids[0] = txtr
        cmdl.material_sets[0].texture_file_ids[1] = emissive
        editor.add_new_asset(f"custom_door_lock_{name}.CMDL", Cmdl(cmdl, Game.ECHOES, editor))


def apply_door_rando(editor: PatcherEditor, world_name: str, area_name: str, dock_name: str,
                     new_door_type: str, old_door_type: str | None, low_memory: bool):
    if old_door_type is not None:
        old_door = DOCK_TYPES[old_door_type]

        if isinstance(old_door, dock_type.VanillaBlastShieldDoorType):
            old_door.remove_blast_shield(editor, world_name, area_name, dock_name)

    new_door = DOCK_TYPES[new_door_type]
    new_door.patch_door(editor, world_name, area_name, dock_name, low_memory)



import struct
from enum import IntEnum
from typing import NamedTuple


class DoorIconColors(NamedTuple):
    surface_color: int
    outline_color: int | None = None  # unused


class DoorMapIcon(IntEnum):
    # vanilla
    Normal = 0
    Missile = 1
    Dark = 2
    Annihilator = 3
    Light = 4
    SuperMissile = 5
    SeekerMissile = 6
    PowerBomb = 7

    # planetary energy
    AgonEnergy = -1
    TorvusEnergy = -2
    SanctuaryEnergy = -3

    # visors
    ScanVisor = -4
    DarkVisor = -5
    EchoVisor = -6

    # misc
    Disabled = -7
    ScrewAttack = -8
    Bomb = -9
    Boost = -10
    Grapple = -11

    @property
    def colors(self) -> DoorIconColors:
        return _ALL_COLORS.get(self, DoorIconColors(0xff00ffff))

    @staticmethod
    def get_surface_colors_as_bytes() -> bytes:
        colors = [icon.colors.surface_color for icon in sorted(DoorMapIcon)]
        return struct.pack(">" + "L" * len(colors), *colors)

    @staticmethod
    def get_door_index_bounds() -> tuple[int, int]:
        return min(*DoorMapIcon), max(*DoorMapIcon)


_ALL_COLORS = {
    DoorMapIcon.Normal: DoorIconColors(0x3379bfff),
    DoorMapIcon.Missile: DoorIconColors(0xd63333ff),
    DoorMapIcon.Dark: DoorIconColors(0x000000ff, 0x898989ff),
    DoorMapIcon.Annihilator: DoorIconColors(0x4b4b4bff),
    DoorMapIcon.Light: DoorIconColors(0xffffffff),
    DoorMapIcon.SuperMissile: DoorIconColors(0x50a148ff),
    DoorMapIcon.SeekerMissile: DoorIconColors(0x794f77ff),
    DoorMapIcon.PowerBomb: DoorIconColors(0xeae50bff),

    DoorMapIcon.AgonEnergy: DoorIconColors(0xa45600ff),
    DoorMapIcon.TorvusEnergy: DoorIconColors(0x4e9761ff),
    DoorMapIcon.SanctuaryEnergy: DoorIconColors(0x56789dff),

    DoorMapIcon.ScanVisor: DoorIconColors(0x007f7fff),
    DoorMapIcon.DarkVisor: DoorIconColors(0x660000ff),
    DoorMapIcon.EchoVisor: DoorIconColors(0xcc7a00ff),

    DoorMapIcon.ScrewAttack: DoorIconColors(0xed94d4ff),
    DoorMapIcon.Bomb: DoorIconColors(0x55557fff),
    DoorMapIcon.Boost: DoorIconColors(0xff5500ff),
    # DoorMapIcon.Grapple: None,

    DoorMapIcon.Disabled: DoorIconColors(0x202020ff),
}



import contextlib
from unittest.mock import MagicMock

import pytest

from open_prime_rando.echoes import dock_lock_rando
from open_prime_rando.echoes.dock_lock_rando import DOCK_TYPES
from open_prime_rando.patcher_editor import PatcherEditor

_custom_asset_ids = {
    'custom_door_lock_darkburst.CMDL': 3171862285,
    'custom_door_lock_darkburst.TXTR': 3672172162,
    'custom_door_lock_darkburst_emissive.TXTR': 234173793,
    'custom_door_lock_echo_visor.CMDL': 2853277507,
    'custom_door_lock_echo_visor.TXTR': 3456085708,
    'custom_door_lock_echo_visor_emissive.TXTR': 1855515304,
    'custom_door_lock_greyscale_emissive.TXTR': 3597936245,
    'custom_door_lock_boost_ball.CMDL': 3448653349,
    'custom_door_lock_boost_ball.TXTR': 2858444714,
    'custom_door_lock_boost_ball_emissive.TXTR': 679049133,
    'custom_door_lock_cannon_ball.CMDL': 3063077393,
    'custom_door_lock_cannon_ball.TXTR': 3514629022,
    'custom_door_lock_cannon_ball_emissive.TXTR': 3792718965,
    'custom_door_lock_charge_beam.CMDL': 3005557644,
    'custom_door_lock_charge_beam.TXTR': 3570076163,
    'custom_door_lock_charge_beam_emissive.TXTR': 1595153749,
    'custom_door_lock_dark_visor.CMDL': 2145020805,
    'custom_door_lock_dark_visor.TXTR': 406080010,
    'custom_door_lock_dark_visor_emissive.TXTR': 775355676,
    'custom_door_lock_morph_ball_bombs.CMDL': 3083075990,
    'custom_door_lock_morph_ball_bombs.TXTR': 3492421657,
    'custom_door_lock_morph_ball_bombs_emissive.TXTR': 805979281,
    'custom_door_lock_screw_attack.CMDL': 1760304179,
    'custom_door_lock_screw_attack.TXTR': 251805116,
    'custom_door_lock_screw_attack_emissive.TXTR': 1723694793,
    'custom_door_lock_sonic_boom.CMDL': 1667093720,
    'custom_door_lock_sonic_boom.TXTR': 78902615,
    'custom_door_lock_sonic_boom_emissive.TXTR': 2606270890,
    'custom_door_lock_sunburst.CMDL': 987291923,
    'custom_door_lock_sunburst.TXTR': 1563869340,
    'custom_door_lock_sunburst_emissive.TXTR': 2638796108,
}


def test_add_custom_models(prime2_editor: PatcherEditor):
    dock_lock_rando.add_custom_models(prime2_editor)

    assert prime2_editor._custom_asset_ids == _custom_asset_ids


vanilla_doors = {
    "Normal": ("Torvus Bog", "Gathering Hall", "North"),
    "Dark": ("Torvus Bog", "Torvus Temple", "EastTop"),
    "Light": ("Torvus Bog", "Underground Tunnel", "North"),
    "Annihilator": ("Torvus Bog", "Gathering Hall", "East_0P"),

    "Missile": ("Temple Grounds", "GFMC Compound", "WestTop"),
    "SuperMissile": ("Torvus Bog", "Torvus Temple", "WestGenerator"),
    "SeekerMissile": ("Temple Grounds", "Path of Honor", "North"),
    "PowerBomb": ("Temple Grounds", "GFMC Compound", "West"),
}


@pytest.mark.parametrize("new_door_type", DOCK_TYPES.keys())
@pytest.mark.parametrize("old_door_type", vanilla_doors.keys())
@pytest.mark.parametrize("low_memory", [False])  # too slow to run twice for now
def test_apply_door_rando(prime2_editor, new_door_type, old_door_type, low_memory):
    prime2_editor.ensure_present = MagicMock()

    if new_door_type in {"Grapple", "AgonEnergy", "TorvusEnergy", "SanctuaryEnergy"}:
        expectation = pytest.raises(NotImplementedError)
    else:
        expectation = contextlib.nullcontext()

    world_name, area_name, dock_name = vanilla_doors[old_door_type]
    with expectation:
        dock_lock_rando.apply_door_rando(prime2_editor, world_name, area_name, dock_name,
                                         new_door_type, old_door_type, low_memory)



import copy

import open_prime_rando.echoes.asset_ids.world
from open_prime_rando.patcher_editor import PatcherEditor


def expand_schema(base_schema: dict, editor: PatcherEditor) -> dict:
    schema = copy.deepcopy(base_schema)

    world_props = schema["properties"]["worlds"]["properties"] = {}
    for world, mlvl_id in open_prime_rando.echoes.asset_ids.world.NAME_TO_ID_MLVL.items():
        world_def = copy.deepcopy(schema["$defs"]["world"])
        world_props[world] = world_def

        mlvl = editor.get_mlvl(mlvl_id)
        area_props = {}
        world_def["properties"]["areas"] = {"type": "object", "additionalProperties": False, "properties": area_props}

        world_details = open_prime_rando.echoes.asset_ids.world.load_dedicated_file(world)
        mrea_to_name: dict[int, str] = {
            mrea: name
            for name, mrea in world_details.NAME_TO_ID_MREA.items()
        }

        for area in mlvl.areas:
            area_name = mrea_to_name[area.mrea_asset_id]
            area_def = copy.deepcopy(schema["$defs"]["area"])
            area_props[area_name] = area_def

            area_def["properties"]["docks"] = {
                "type": "object",
                "properties": {
                    dock_name: {"$ref": "#/$defs/dock"}
                    for dock_name in world_details.DOCK_NAMES.get(area_name, {}).keys()
                },
                "default": {},
                "additionalProperties": False,
            }

            area_def["properties"]["layers"] = {
                "type": "object",
                "properties": {
                    layer.name: {"type": "boolean"}
                    for layer in area.layers
                },
                "default": {},
                "additionalProperties": False,
            }

    return schema



from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.properties.echoes.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.echoes.objects.Timer import Timer

import open_prime_rando.echoes.asset_ids.agon_wastes as agon_wastes
import open_prime_rando.echoes.asset_ids.great_temple as great_temple
import open_prime_rando.echoes.asset_ids.sanctuary_fortress as sanctuary_fortress
import open_prime_rando.echoes.asset_ids.temple_grounds as temple_grounds
import open_prime_rando.echoes.asset_ids.torvus_bog as torvus_bog
import open_prime_rando.echoes.asset_ids.world as world
from open_prime_rando.patcher_editor import PatcherEditor

ELEVATOR_MEMORY_RELAY_PER_MREA = {
    world.GREAT_TEMPLE_MLVL: {
        great_temple.TEMPLE_TRANSPORT_A_MREA: 0x00000107,
        great_temple.TEMPLE_TRANSPORT_B_MREA: 0x0008002f,
        great_temple.TEMPLE_TRANSPORT_C_MREA: 0x00060046,
    },

    world.TEMPLE_GROUNDS_MLVL: {
        temple_grounds.TRANSPORT_TO_AGON_WASTES_MREA: 0x00180040,
        temple_grounds.TRANSPORT_TO_SANCTUARY_FORTRESS_MREA: 0x00330058,
        temple_grounds.TRANSPORT_TO_TORVUS_BOG_MREA: 0x001e0061,
    },

    world.AGON_WASTES_MLVL: {
        agon_wastes.TRANSPORT_TO_SANCTUARY_FORTRESS_MREA: 0x002d0064,
        agon_wastes.TRANSPORT_TO_TEMPLE_GROUNDS_MREA: 0x0000003f,
        agon_wastes.TRANSPORT_TO_TORVUS_BOG_MREA: 0x0013004d,
    },

    world.TORVUS_BOG_MLVL: {
        torvus_bog.TRANSPORT_TO_AGON_WASTES_MREA: 0x00210038,
        torvus_bog.TRANSPORT_TO_SANCTUARY_FORTRESS_MREA: 0x0045004b,
        torvus_bog.TRANSPORT_TO_TEMPLE_GROUNDS_MREA: 0x0000002e,
    },

    world.SANCTUARY_FORTRESS_MLVL: {
        sanctuary_fortress.TRANSPORT_TO_AGON_WASTES_MREA: 0x00130088,
        sanctuary_fortress.TRANSPORT_TO_TEMPLE_GROUNDS_MREA: 0x00000035,
        sanctuary_fortress.TRANSPORT_TO_TORVUS_BOG_MREA: 0x00190036,
    }
}


def apply_auto_enabled_elevators_patch(editor: PatcherEditor):
    """
    Patches that activates every elevator on room load
    """
    for mlvl_id, areas in ELEVATOR_MEMORY_RELAY_PER_MREA.items():
        for mrea_id, memory_relay_id in areas.items():
            area = editor.get_area(mlvl_id, mrea_id)
            timer = area.get_layer("Default").add_instance_with(Timer(
                editor_properties=EditorProperties(
                    name="Timer - Auto enable elevator",
                    active=False
                ),
                time=0.001,
                auto_start=True
            ))
            relay = area.get_instance(memory_relay_id)
            timer.add_connection(State.Zero, Message.Activate, relay)



import hashlib
from pathlib import Path

import pytest

from open_prime_rando import echoes_patcher


@pytest.fixture(
    params=[False, True],
)
def is_legacy(request: pytest.FixtureRequest) -> bool:
    return request.param


def hash_all_paks(base_path: Path) -> dict[str, bytes]:
    return {
        pak_path.relative_to(base_path).as_posix(): hashlib.sha256(pak_path.read_bytes()).digest()
        for pak_path in base_path.rglob("*.pak")
    }


def test_ntsc_paks(prime2_iso_provider, tmp_path, test_files_dir, is_legacy: bool) -> None:
    output_path = tmp_path.joinpath("out")
    configuration = test_files_dir.read_json("echoes", "door_lock.json")
    configuration["legacy_compatibility"] = is_legacy

    expected_hashes = {
        False: {
            'AudioGrp.pak': b'U\xc6\xfe\\\xac\xcd\xdc\x02\x0c\xed\xdd\xb3\xfarbP'
                            b'\x90\xac\x8d\x1b=\xd4\x9fl\x82\xc8\x12\xd2\x85X{+',
            'LogBook.pak': b'\x05\xdf2\xb5\xf4\xe5\xce\xa3"\tk_\x91\x05\xe9\xa9'
                           b'\xab\x8c.\x9c\xe0\xee\x89\xb7\x04\xa1\x7fI\x0bU,\xb4',
            'Metroid1.pak': b'\x11%\x80\xe4t\x9e?\x1d\xaa\xf2\xc9\xc3\x1d&\xa6v\xfc?E\xcf'
                            b'\xc4\x06\x13\x83"A#\xfb\xa1\x84\xcf\xca',
            'Metroid2.pak': b'\x96lH\x85\xae\xddX\xceo\xd1\x0cfnk\xbex\xaa\x94Y\x08'
                            b'\x05\\-\xf9?\xa4\x9a]\xa8\x0ez\x93',
            'Metroid3.pak': b'\xe8\xee\x80\xf6\\\xd5\x1c4\xf5\x1b\xe1*\xfbY\x9a\xe0'
                            b'\x8c\xb9\xd1\xe0(\x96\xb4R-\xe0\xf1\x86y\x88T\xa9',
            'Metroid4.pak': b'\xce\xc5\x838\x19q\xd0\xf7\xf4\x86hQ\xb3\xf5\xf5p\xf9\xd3}8'
                            b'\xfemx\xb5x\tQ\xf8\xc6\xdc.\xb6',
            'Metroid5.pak': b'\xd0\x9fX\xab\xae\x06\xda\xbfz\xfd\xc5\xc9\xcb+\x0f\x04'
                            b'\xd7\x9c\xa0\xb7\xd4\xf4\xea\x85\x14\xa4\x08\xe5'
                            b'\xf6\x9cY\x11',
            'MiscData.pak': b'\x0bM\xedR\xb7+\xaf+#\x06\x87\x88\xe5\xec\xbbxg@i\xbc'
                            b'\xa3n\xb1\xabi\xfajb\x00\xf3\xd1\xaa',
            'SamusGun.pak': b'\xdd\xa99\x0b\x14b\x97&\xee\xdf\nE\xd0GR`c\x85\xf7m;\xdbX^'
                            b'\xe4\x87;\\?\xc9\x08\x1f',
            'SamusGunLow.pak': b'o8\xb8\xab^\xc3y\xc7\xd8v\xf1?\xe7a\x9enIE\xb7\xa5'
                               b'q\x9f\xa2\xf6\xc2\xc6m&\x80G\x8f\x83',
            'TestAnim.pak': b'\xc0n\xfa2\x87\x96o\xd9\xf8\nL\xc8\xd5vVJi\xe2\xa0\xdb'
                            b'\xdbq\t\x7f\xaf\x96\x8d\x02\x1f\xe7\x07\x12'
        },
        True: {
            'AudioGrp.pak': b'U\xc6\xfe\\\xac\xcd\xdc\x02\x0c\xed\xdd\xb3\xfarbP'
                            b'\x90\xac\x8d\x1b=\xd4\x9fl\x82\xc8\x12\xd2\x85X{+',
            'LogBook.pak': b'K=z\x02\x9c>\x8a\xfbN\xbe\xb6\xb5\xc8\xad\xba\xb8'
                           b'\x99\xae\x85_\xab55\xfe3\x15\x0b\xdc\xd1\xc7\x02^',
            'Metroid1.pak': b'7\x18\xe7+\xdbR\xbf6\x0e\xa2=\x8b\xc8\xcc\x98-'
                            b'\x8f\xdc\xe5\x8d\x17\xedz\xf9c\xe1\xd5\xfc\xf2\x18\xf2$',
            'Metroid2.pak': b'U\x9b\xa7;\xbf\xad\xac\xf7\x05\xe9hKC\xcd\xa5\x0f'
                            b'\xc9C\xc7\xf2\xe7A$\xcc\x1c&0A\xcb\xaa(\xf1',
            'Metroid3.pak': b'\xeaO\x17\x0e\xaa\xbd\xaa\xabJvO\x9d\x08\x15Y\xca'
                            b'\xe7\x97z\x0c\xcb\xf6],\xf9\xfa\xee*\x17\xc2\xf6(',
            'Metroid4.pak': b'9\xc2\x87\x91\x91%\x82\x0f>\x7f;Q\xacLP\xca-\xb6\xc1\x1b'
                            b'+\xc7\xecy{\xc0\x12\x19\xa7\x04\x7ft',
            'Metroid5.pak': b'g\x91Rx\xb0+\xe2=\x06~\xbb\\\x15\xae3\xaft\xdcG\xc2'
                            b'+\xea0\xbf\x8a\xaf\xfc/5`0z',
            'MiscData.pak': b'\x0bM\xedR\xb7+\xaf+#\x06\x87\x88\xe5\xec\xbbxg@i\xbc'
                            b'\xa3n\xb1\xabi\xfajb\x00\xf3\xd1\xaa',
            'SamusGun.pak': b'\xdd\xa99\x0b\x14b\x97&\xee\xdf\nE\xd0GR`c\x85\xf7m;\xdbX^'
                            b'\xe4\x87;\\?\xc9\x08\x1f',
            'SamusGunLow.pak': b'o8\xb8\xab^\xc3y\xc7\xd8v\xf1?\xe7a\x9enIE\xb7\xa5'
                               b'q\x9f\xa2\xf6\xc2\xc6m&\x80G\x8f\x83',
            'TestAnim.pak': b'\xc0n\xfa2\x87\x96o\xd9\xf8\nL\xc8\xd5vVJi\xe2\xa0\xdb'
                            b'\xdbq\t\x7f\xaf\x96\x8d\x02\x1f\xe7\x07\x12'
        },
    }

    echoes_patcher.patch_paks(
        file_provider=prime2_iso_provider,
        output_path=output_path,
        configuration=configuration,
    )
    hashes = hash_all_paks(output_path)
    assert hashes == expected_hashes[is_legacy]


def test_pal_paks(pal_prime2_iso_provider, tmp_path, test_files_dir):
    output_path = tmp_path.joinpath("out")
    configuration = test_files_dir.read_json("echoes", "door_lock.json")

    echoes_patcher.patch_paks(
        file_provider=pal_prime2_iso_provider,
        output_path=output_path,
        configuration=configuration,
    )
    assert len(list(output_path.rglob("*.pak"))) == 11



import argparse
import os.path
import shutil
from pathlib import Path

import retro_data_structures.exceptions
from retro_data_structures.asset_manager import IsoFileProvider
from retro_data_structures.formats import Mrea, Strg
from retro_data_structures.game_check import Game
from retro_data_structures.properties.shared_objects import Dock

from open_prime_rando.patcher_editor import PatcherEditor
from open_prime_rando.unique_area_name import CUSTOM_AREA_NAMES

_CUSTOM_WORLD_NAMES = {
    Game.ECHOES: {
        0x69802220: None,  # "FrontEnd",
        0x7B6EAA68: None,  # "FrontEnd",
        0xA50A80CC: None,  # "M01_SidehopperStation",
        0xAE171602: None,  # "M02_Spires",
        0xE3B0C703: None,  # "M03_CrossfireChaos",
        0x233E42BE: None,  # "M04_Pipeline",
        0x406ADD7F: None,  # "M05_SpiderComplex",
        0x7E19ED26: None,  # "M06_ShootingGallery",
    }
}
_CUSTOM_AREA_NAMES = {
    Game.ECHOES: CUSTOM_AREA_NAMES,
}


# Complexity
# ruff: noqa: C901

def filter_name(s: str) -> str:
    result = s.replace("!", "").replace(" ", "_").replace("'", "").replace(
        '"', "").replace("(", "").replace(")", "").upper()
    while result and not result[0].isalpha():
        result = result[1:]
    return result


def filter_and_lower_name(s: str) -> str:
    return filter_name(s).lower()


def dock_name_templates(dock_names: dict[str, dict[str, int]]) -> str:
    template = "\nDOCK_NAMES = {\n"
    for name in sorted(dock_names.keys()):
        template += f"    \"{name}\": {{\n"
        for dock_name, dock_number in sorted(dock_names[name].items(), key=lambda it: it[1]):
            template += f"        \"{dock_name}\": {dock_number},\n"
        template += "    },\n"
    template += "}\n"

    return template


def generate_template(items: dict[str, int], suffix: str) -> str:
    template = f"# Generated by {os.path.basename(__file__)}\n\n"
    template += "\n".join(
        f"{filter_name(key)}{suffix} = 0x{items[key]:08X}"
        for key in sorted(items)
    )
    template += "\n"

    template += f"\nNAME_TO_ID{suffix}" + " = {\n"
    for name in sorted(items):
        template += f"    \"{name}\": 0x{items[name]:08X},\n"
    template += "}\n"

    return template


def create_asset_id_files(editor: PatcherEditor, output_path: Path):
    output_path.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(output_path)
    output_path.mkdir()

    custom_world_names = _CUSTOM_WORLD_NAMES.get(editor.target_game, {})
    world_names = {}
    mapw_names = {}

    for value in editor.all_asset_ids():
        if editor.get_asset_type(value).lower() != "mlvl":
            continue

        mlvl = editor.get_mlvl(value)

        try:
            strg = editor.get_file(mlvl.raw.world_name_id, type_hint=Strg)
            world_name = strg.raw.string_tables[0].strings[0].string
        except retro_data_structures.exceptions.UnknownAssetId:
            if value not in custom_world_names:
                raise

            world_name = custom_world_names[value]
            if world_name is None:
                continue

        world_names[world_name] = value
        mapw_names[world_name] = mlvl.raw.world_map_id

        area_names = {}
        mapa_names = {}
        dock_names = {}
        area_id_names = {}

        for area, mapa in zip(mlvl.raw.areas, mlvl.mapw.mapa_ids):
            try:
                strg = editor.get_file(area.area_name_id, type_hint=Strg)
                area_name = strg.raw.string_tables[0].strings[0].string
            except retro_data_structures.exceptions.UnknownAssetId:
                area_name = area.internal_area_name
            area_name = _CUSTOM_AREA_NAMES[editor.target_game].get(area.area_mrea_id, area_name)

            if area_name in area_names:
                area_name += "_2"
            assert area_name not in area_names, area_name
            area_names[area_name] = area.area_mrea_id
            mapa_names[area_name] = mapa
            area_id_names[area_name] = area.internal_area_id
            mrea = editor.get_file(area_names[area_name], type_hint=Mrea)

            docks = {}
            for layer in mrea.script_layers:
                for obj in layer.instances:
                    if obj.type_name == "DOCK":
                        dock: Dock = obj.get_properties_as(Dock)
                        assert dock.get_name() not in docks
                        docks[dock.get_name()] = dock.dock_number

                # Docks are in the default layer, ignore the rest
                break

            assert max(docks.values(), default=-1) == len(docks) - 1
            dock_names[area_name] = docks

        world_file_body = generate_template(area_names, "_MREA")
        world_file_body += generate_template(mapa_names, "_MAPA")
        world_file_body += generate_template(area_id_names, "_INTERNAL_ID")
        world_file_body += dock_name_templates(dock_names)
        output_path.joinpath(f"{filter_and_lower_name(world_name)}.py").write_text(world_file_body)

    global_file_body = generate_template(world_names, "_MLVL")
    global_file_body += generate_template(mapw_names, "_MAPW")

    global_file_body += "\n_DEDICATED_FILES = {\n"
    for name in sorted(world_names.keys()):
        global_file_body += f"    \"{name}\": \".{filter_and_lower_name(name)}\",\n"
    global_file_body += """}


def load_dedicated_file(world_name: str):
    import importlib
    return importlib.import_module(
        _DEDICATED_FILES[world_name],
        ".".join(__name__.split(".")[:-1]),
    )
"""

    output_path.joinpath("world.py").write_text(global_file_body)

    all_imports = list(map(filter_and_lower_name, world_names.keys()))
    all_imports.append("world")

    output_path.joinpath("__init__.py").write_text("\n".join([
        "from . import (\n    " + ",\n    ".join(all_imports),
        ")",
        "",
        "__all__ = [",
        *[f'    "{it}",' for it in all_imports],
        "]",
        "",
    ]))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--game", required=True, choices=["echoes"])
    parser.add_argument("--iso", required=True, type=Path,
                        help="Path to where the ISO.")
    args = parser.parse_args()

    create_asset_id_files(
        PatcherEditor(IsoFileProvider(args.iso), Game.ECHOES),
        Path(__file__).parents[1].joinpath("src", "open_prime_rando", args.game, "asset_ids"),
    )


if __name__ == '__main__':
    main()



import dataclasses
import functools
from pathlib import Path

from retro_data_structures.asset_manager import AssetManager
from retro_data_structures.base_resource import RawResource
from retro_data_structures.formats import Ancs, Cmdl


@functools.cache
def custom_asset_path() -> Path:
    return Path(__file__).parent


@dataclasses.dataclass
class TranslatorAssets:
    name: str
    txtr: int
    new_cmdl: int
    new_ancs: int


SCAN_VISOR_CMDL = 0xafc70004
DARK_VISOR_ANCS = 0x851b526e

VIOLET_TXTR = 0x4be5342e
AMBER_TXTR = 0xf5308558
EMERALD_TXTR = 0xa9640fdf
COBALT_TXTR = 0x2c56d2d4

VIOLET_TRANSLATOR = TranslatorAssets(
    name="violet_translator",
    txtr=0x4be5342e,
    new_cmdl=1448365380,
    new_ancs=1448362318,
)
AMBER_TRANSLATOR = TranslatorAssets(
    name="amber_translator",
    txtr=0xf5308558,
    new_cmdl=1096043844,
    new_ancs=1096040782,
)
EMERALD_TRANSLATOR = TranslatorAssets(
    name="emerald_translator",
    txtr=0xa9640fdf,
    new_cmdl=1163152708,
    new_ancs=1163149646,
)
COBALT_TRANSLATOR = TranslatorAssets(
    name="cobalt_translator",
    txtr=0x2c56d2d4,
    new_cmdl=1129598276,
    new_ancs=1129595214,
)


def _create_visor_derivatives(editor: AssetManager):
    for translator in [VIOLET_TRANSLATOR, AMBER_TRANSLATOR, EMERALD_TRANSLATOR, COBALT_TRANSLATOR]:
        editor.register_custom_asset_name(f"{translator.name}_cmdl", translator.new_cmdl)
        editor.register_custom_asset_name(f"{translator.name}_ancs", translator.new_ancs)

        cmdl = editor.get_parsed_asset(SCAN_VISOR_CMDL, type_hint=Cmdl)
        cmdl.raw.material_sets[0].texture_file_ids[0] = translator.txtr
        cmdl.raw.material_sets[0].texture_file_ids[1] = translator.txtr
        editor.add_new_asset(f"{translator.name}_cmdl", cmdl, editor.find_paks(SCAN_VISOR_CMDL))

        ancs = editor.get_parsed_asset(DARK_VISOR_ANCS, type_hint=Ancs)
        ancs.raw.character_set.characters[0].model_id = translator.new_cmdl
        editor.add_new_asset(f"{translator.name}_ancs", ancs, editor.find_paks(DARK_VISOR_ANCS))

    editor.register_custom_asset_name("scan_visor_ancs", editor.generate_asset_id())
    ancs = editor.get_parsed_asset(DARK_VISOR_ANCS, type_hint=Ancs)
    ancs.raw.character_set.characters[0].model_id = SCAN_VISOR_CMDL
    editor.add_new_asset("scan_visor_ancs", ancs, editor.find_paks(DARK_VISOR_ANCS))


@dataclasses.dataclass
class BeamAmmoAssets:
    name: str
    txtr_a: int
    txtr_b: int
    particle: int
    new_cmdl: int
    new_ancs: int


BEAM_AMMO_EXPANSION_CMDL = 0x352c8b02
BEAM_AMMO_EXPANSION_ANCS = 0x4e00188c

DARK_AMMO = BeamAmmoAssets(
    name="dark_ammo",
    txtr_a=1385637646,
    txtr_b=3277274054,
    particle=0x8e0c499d,
    new_cmdl=1631670273,
    new_ancs=0x61415002,
)
LIGHT_AMMO = BeamAmmoAssets(
    name="light_ammo",
    txtr_a=1815959726,
    txtr_b=2738959665,
    particle=0xa180bb7f,
    new_cmdl=1631670275,
    new_ancs=1631670276,
)


def _create_split_ammo(editor: AssetManager):
    for ammo in [DARK_AMMO, LIGHT_AMMO]:
        editor.register_custom_asset_name(f"{ammo.name}_cmdl", ammo.new_cmdl)
        editor.register_custom_asset_name(f"{ammo.name}_ancs", ammo.new_ancs)

        cmdl = editor.get_parsed_asset(BEAM_AMMO_EXPANSION_CMDL, type_hint=Cmdl)
        cmdl.raw.material_sets[0].texture_file_ids[0] = ammo.txtr_a
        cmdl.raw.material_sets[0].texture_file_ids[2] = ammo.txtr_a
        cmdl.raw.material_sets[0].texture_file_ids[3] = ammo.txtr_b
        cmdl.raw.material_sets[0].texture_file_ids[4] = ammo.txtr_b
        editor.add_new_asset(f"{ammo.name}_cmdl", cmdl, editor.find_paks(BEAM_AMMO_EXPANSION_CMDL))

        ancs = editor.get_parsed_asset(BEAM_AMMO_EXPANSION_ANCS, type_hint=Ancs)
        ancs.raw.character_set.characters[0].model_id = ammo.new_cmdl
        ancs.raw.character_set.characters[0].particle_resource_data.generic_particles = [
            0x89C2F268,
            ammo.particle,
        ]
        ancs.raw.animation_set.event_sets[0].particle_poi_nodes[0].particle.id = ammo.particle
        ancs.raw.animation_set.event_sets[0].particle_poi_nodes[1].particle.id = ammo.particle
        editor.add_new_asset(f"{ammo.name}_ancs", ancs, editor.find_paks(BEAM_AMMO_EXPANSION_ANCS))


def _import_premade_assets(editor: AssetManager):
    assets = Path(__file__).parent.joinpath("custom_assets", "general")
    for f in assets.glob("*.*"):
        name = f.name
        asset_type = f.suffix[1:].upper() # remove leading period, force uppercase
        raw = f.read_bytes()
        editor.add_new_asset(name, RawResource(asset_type, raw), ())


def create_custom_assets(editor: AssetManager, include_premade: bool = False):
    _create_visor_derivatives(editor)
    _create_split_ammo(editor)
    _import_premade_assets(editor)



import random

from retro_data_structures.enums.echoes import Message, ScanSpeed, State
from retro_data_structures.formats.scan import Scan
from retro_data_structures.formats.strg import Strg
from retro_data_structures.properties.echoes.objects.PointOfInterest import PointOfInterest
from retro_data_structures.properties.echoes.objects.ScannableObjectInfo import ScannableObjectInfo
from retro_data_structures.properties.echoes.objects.Sound import Sound
from retro_data_structures.properties.echoes.objects.Switch import Switch

from open_prime_rando.echoes.asset_ids.sanctuary_fortress import MAIN_GYRO_CHAMBER_MREA, SENTINELS_PATH_MREA
from open_prime_rando.echoes.asset_ids.temple_grounds import PROFANE_PATH_MREA
from open_prime_rando.echoes.asset_ids.world import SANCTUARY_FORTRESS_MLVL, TEMPLE_GROUNDS_MLVL
from open_prime_rando.patcher_editor import PatcherEditor

ECHO_LOCK_MREAS = [
    (SANCTUARY_FORTRESS_MLVL, MAIN_GYRO_CHAMBER_MREA),
    (SANCTUARY_FORTRESS_MLVL, SENTINELS_PATH_MREA),
    (TEMPLE_GROUNDS_MLVL, PROFANE_PATH_MREA)
]
ECHO_LOCK_STATES = [State.Zero, State.InternalState00, State.InternalState01, State.InternalState02]
ECHO_LOCK_SOUNDS = [1005, 1006, 1007]


def randomize_echo_locks(editor: PatcherEditor, rng: random.Random):
    # create key scan assets
    key_scan = editor.get_parsed_asset(0x2E7C4349, type_hint=Scan)
    scan_info = key_scan.scannable_object_info.get_properties_as(ScannableObjectInfo)
    key_strg = editor.get_parsed_asset(0x43894960, type_hint=Strg)

    key_scans = []
    for pitch in ["low", "medium", "high"]:
        key_strg.set_string(
            1,
            "Sonic detection gear needed to interface with this system."
            " Shoot the Echo Key Beam emitter with a sonic pulse to activate it."
            f" It will then fire a blast of &push;&main-color=#FF3333;{pitch}-pitched&pop; sound at an Echo Gate lock."
        )
        strg_id = editor.add_new_asset(f"accessible_echo_lock_{pitch}.STRG", key_strg)

        scan_info.string = strg_id
        key_scan.scannable_object_info.set_properties(scan_info)
        key_scan.rebuild_dependencies()
        scan_id = editor.add_new_asset(f"accessible_echo_lock_{pitch}.SCAN", key_scan)

        key_scans.append(scan_id)

    gate_scan = editor.get_parsed_asset(0x80A987AA, type_hint=Scan)
    gate_scan_info = gate_scan.scannable_object_info.get_properties_as(ScannableObjectInfo)
    gate_scan_info.scan_speed = ScanSpeed.Slow
    gate_strg = editor.get_parsed_asset(0x2820CC3D, type_hint=Strg)

    for mlvl_id, mrea_id in ECHO_LOCK_MREAS:
        area = editor.get_area(mlvl_id, mrea_id)
        solution = [rng.randint(0, 2) for _ in range(4)]

        counter = area.get_instance("Lock Solution Counter")
        gate_poi = area.get_instance("(Gate) Echo Gate Scan Point")
        correct_key_relays = [area.get_instance(f"[IN] Set Key {i} As Correct Choice") for i in range(3)]
        lock_tone_players = [area.get_instance(f"(Gate) Lock {i} Tone") for i in range(4)]
        key_switches = [area.get_instance(f"(Key {i}) Check Echo Key") for i in range(3)]
        key_scan_points = [area.get_instance(f"(Key {i}) Scan Point") for i in range(3)]

        for relay in correct_key_relays:
            counter.remove_connections_from(relay)

        # Update all scan posts to have the updated scan text
        for scan_point, key_scan in zip(key_scan_points, key_scans):
            with scan_point.edit_properties(PointOfInterest) as scan:
                scan.scan_info.scannable_info0 = key_scan

        for i, key in enumerate(solution):
            counter.add_connection(ECHO_LOCK_STATES[i], Message.SetToZero, correct_key_relays[key])

            with lock_tone_players[i].edit_properties(Sound) as tone:
                tone.sound = ECHO_LOCK_SOUNDS[key]

        for i, switch in enumerate(key_switches):
            with switch.edit_properties(Switch) as props:
                props.is_open = i == solution[0]

        # edit scan to indicate the solution
        solution_text = ("Sonic detection gear needed to interface with this system. "
                         "The combination of its sonic locks is:\n")
        solution_text += ", ".join(["Low", "Medium", "High"][key] for key in solution)
        gate_strg.set_string(1, solution_text)
        strg_id = editor.add_new_asset(f"accessible_echo_gate_{mrea_id}.STRG", gate_strg)

        gate_scan_info.string = strg_id
        gate_scan.scannable_object_info.set_properties(gate_scan_info)
        scan_id = editor.add_new_asset(f"accessible_echo_gate_{mrea_id}.SCAN", gate_scan)

        with gate_poi.edit_properties(PointOfInterest) as poi:
            poi.scan_info.scannable_info0 = scan_id

        area.update_all_dependencies()




TG_PAIRS = [
    # 01_temple_landingsite (Landing Site / Defiled Shrine)
    (0x62B0D67D, 0x16E116B4),
    # 0q_temple_hall (Service Access / Shrine Access)
    (0x9FB7503C, 0x1F1D07DF),
    # 0t_temple_hall (Path of Honor / Gateway Access)
    (0x7B4D31DB, 0x84324C98),
    # 09_temple_thirdleg (Meeting Grounds / War Ritual Grounds)
    (0x9B5EFA75, 0xAA77EAAD),
    # 12_temple_finalgate (Hall of Honored Dead / Sky Temple Gateway)
    (0xB8B34634, 0x87D35EE4),
    # 0n_temple_hall (Hall of Eyes / Base Access)
    (0x476BA84C, 0x686F5A1B),
    # 02_temple_exploration (Industrial Site / Accursed Lake)
    (0x62AC8AC4, 0x9FD196DF),
    # 10_temple_pathofeyes (Path of Eyes / Abandoned Base)
    (0xEE4732BE, 0xC6F4E0C2),
    # 0b_temple_hall (Collapsed Tunnel / Lake Access)
    (0x5B92D146, 0x688AB8AE),
    # 01_temple_hive03 (Hive Chamber B / Portal Site)
    (0x1D7BD3AE, 0xDA37FF97),
    # 03_temple_combattraining (Temple Assembly Site / Plain of Dark Worship)
    (0x4B4E5911, 0x0405EBA4),
    # 11_temple_grandpuzzle (Grand Windchamber / Ing Windchamber)
    (0xAD70917D, 0xCDEF2BF1),
    # 04_temple_unknownenemies (Communication Area / Reliquary Grounds)
    (0x635E8DA1, 0xFC8FC0A4),
    # 05_temple_federationsite (GFMC Compound / Phazon Grounds)
    (0x57757AD5, 0x645EF243),
    # 00_pickup_temple_a (Storage Cavern A / Ing Reliquary)
    (0x3800CF82, 0x749E82E3),
    # 0d_temple_hall (Trooper Security Station / Reliquary Access)
    (0x55EE6DC3, 0x8540CAD4),
    # 0e_temple_hall (Sacred Bridge / Phazon Pit)
    (0xBABCDB22, 0x1E332000),
    # 06_temple_sacredpath (Sacred Path / Profane Path)
    (0xC23CFFBD, 0x64E640D6),
]

AGON_PAIRS = [
    # 0a_sand_hall (Plaza Access / Ing Cache 4)
    (0xADF88927, 0xC5C7C9D3),
    # 01_sand_entrance (Mining Plaza / Duelling Range)
    (0x427FAD9A, 0x3F5CDF39),
    # 0e_sand_hall (Transit Station / Dark Transit Station)
    (0x79C9C5A0, 0xD73ECD68),
    # 0f_sand_hall (Save Station A / Save Station 2)
    (0x50017152, 0xAC204F8B),
    # 0o_sand_hall (Mining Station Access / Junction Site)
    (0x8D7A7F8C, 0x1608C563),
    # 04_sand_mining (Mining Station B / Trial Grounds)
    (0xDB7B2CED, 0x86C1AB37),
    # 0q_sand_hall (Transport Center / Crossroads)
    (0x4BDFB7B9, 0x8E23DB3F),
    # 02_sand_sandfalls (Mining Station A / Judgment Pit)
    (0xE3BEF27F, 0x6B77F9BD),
    # 0d_sand_hall (Mine Shaft / Trial Tunnel)
    (0xD7A15431, 0x48E44EF6),
    # 05_portalpuzzle (Portal Terminal / Portal Site)
    (0x2BCD44A7, 0xBF7F572C),
    # 0c_sand_hall (Temple Access / Dark Agon Temple Access)
    (0x2A58AC44, 0x2103C8AE),
    # 0g_sand_hall (Central Station Access / Warrior's Walk)
    (0xFE69E0C3, 0x33FACC15),
    # 00_save_sand_a (Sand Cache / Save Station 1)
    (0xF4B1F727, 0x5A154E2E),
    # 0n_sand_hall (Portal Access A / Portal Access)
    (0x2312EE1D, 0x89D246FD),
    # 03_sand_temple (Agon Temple / Dark Agon Temple)
    (0x75FC9AAE, 0x7575048A),
    # 06_sand_alamofront (Central Mining Station / Battleground)
    (0xF5A6D972, 0xEA79522C),
    # 00_sand_generator_connect (Controller Access / Dark Controller Access)
    (0xBF5399C0, 0xA2F347D1),
    # 0h_sand_hall (Command Center Access / Double Path)
    (0x708387F9, 0x7FEF433B),
    # 00_sand_generator (Agon Energy Controller / Dark Agon Energy Controller)
    (0x02FC3717, 0xB7DD4006),
    # 07_sand_alamoportalobservation (Command Center / Doomed Entry)
    (0x31C25557, 0xA49F8A6D),
    # 10_sand_alamostorage (Main Reactor / Dark Oasis)
    (0xCCD9F39E, 0x99FB957F),
    # 0i_sand_hall (Biostorage Access / Feeding Pit Access)
    (0xDEEB1668, 0xE035C0A5),
    # 0m_sand_hall (Security Station B / Oasis Access)
    (0x0ADA5AEF, 0xF2CCC41E),
    # 0l_sand_hall (Sand Processing / Hall of Stairs)
    (0xA4B2CB7E, 0x6D164780),
    # 00_pickup_sand_d (Storage D / Ing Cache 3)
    (0x40FFA8F7, 0xBA23E5F5),
    # 08_sand_casualpirates (Biostorage Station / Feeding Pit)
    (0x80896CE9, 0x075A76EB),
    # 00_save_sand_c (Save Station C / Save Station 3)
    (0xB979562C, 0x0279F7EF),
    # 0j_sand_hall (Security Station A / Watering Hole)
    (0xF723A29A, 0x9B2B4246),
    # 00_pickup_sand_b (Storage B / Ing Cache 1)
    (0x96A64BEA, 0x52962FB6),
    # 0k_sand_hall (Ventilation Area B / Bitter Well)
    (0x594B330B, 0x04F1C1D8),
    # 09_sand_testtubepuzzle (Bioenergy Production / Phazon Site)
    (0x13470DA8, 0x200EF5E7),
    # 00_pickup_sand_c (Storage C / Ing Cache 2)
    (0x5DFA984F, 0x9318F076),
]

TORVUS_PAIRS = [
    # 01_swamp_entrance (Torvus Lagoon / Poisoned Bog)
    (0x7E0BE699, 0x14E2C3A6),
    # 0d_swamp_hall (Ruined Alcove / Putrid Alcove)
    (0xD37C2414, 0xCBC72D79),
    # 0c_swamp_hall (Portal Chamber / Portal Chamber)
    (0xF3EE585F, 0xAE1E1339),
    # 00_pickup_swamp_a (Save Station A / Cache A)
    (0xDB6E4075, 0x0CAAEEC6),
    # 03_swamp_forgottenbridge (Forgotten Bridge / Dark Forgotten Bridge)
    (0x09DECF21, 0x0253782D),
    # 02_swamp_greatbridge (Great Bridge / Venomous Pond)
    (0xE3D5A55E, 0x858EECD3),
    # 0i_swamp_hall (Plaza Access / Dark Arena Tunnel)
    (0xC1C28706, 0x390DFCB9),
    # 0e_swamp_hall (Abandoned Worksite / Brooding Ground)
    (0x54DAEF57, 0xDCEF49B9),
    # 0h_swamp_hall (Grove Access / Dark Falls)
    (0x46644C45, 0x2E259879),
    # 0f_swamp_hall (Temple Access / Dark Torvus Temple Access)
    (0x0740B4D3, 0xE597E4F9),
    # 00_map_swamp (Torvus Map Station / Save Station 1)
    (0x282B587F, 0x9A777442),
    # 06_swamp_boostballguardian (Torvus Plaza / Dark Torvus Arena)
    (0x75F158FF, 0xA49A572F),
    # 05_swamp_timber! (Torvus Grove / Polluted Mire)
    (0xE838FA18, 0x839DC4F6),
    # 04_swamp_temple (Torvus Temple / Dark Torvus Temple)
    (0xE640CE96, 0x3663B195),
    # 00_pickup_swamp_b (Meditation Vista / Gloom Vista)
    (0x5DFA32DB, 0x954888C7),
    # 00_elev_swamp_sand_e (Transport to Agon Wastes / Ammo Station)
    (0xCF659F4E, 0x8697F62A),
    # 0j_swamp_hall (Underground Transport / Cache B)
    (0x9258DC82, 0x007551F9),
    # 00_swamp_genconnect (Controller Access / Dark Controller Access)
    (0x2AF65B49, 0x11C4526B),
    # 07_swamp_shaft (Hydrodynamo Station / Undertemple Shaft)
    (0xD5B5B066, 0xB2D3E26C),
    # 00_swamp_generator (Torvus Energy Controller / Dark Torvus Energy Controller)
    (0x133BF5B8, 0x5ED7CD19),
    # 0l_swamphall (Gathering Access / Crypt Tunnel)
    (0x530DFF81, 0x4DC421AD),
    # ok_swamp_hall (Training Access / Sacrificial Chamber Tunnel)
    (0x77AF9FD1, 0x65E571CD),
    # 00_save_swamp_b (Save Station B / Save Station 2)
    (0xCB165BD8, 0x13254EC6),
    # 0q_swamp_hall (Hydrodynamo Shaft / Undertemple Access)
    (0x308351E5, 0x29B990F8),
    # 10_swamp_hiddenhalfpipe (Gathering Hall / Crypt)
    (0x3501473A, 0xC8A84B98),
    # 08_swamp_grappleguardian (Training Chamber / Sacrificial Chamber)
    (0x4BB5AE60, 0x629790F4),
    # 09_swamp_catacombs (Catacombs / Dungeon)
    (0xFB628DCB, 0xBBE4B3AE),
    # 11_swamp_bloggtank (Main Hydrochamber / Undertemple)
    (0xCEBAC0BD, 0x85A3801F),
    # 0n_swamp_hall (Transit Tunnel West / Undertransit One)
    (0xE150FB4D, 0x5CD4C2F9),
    # 0o_swamp_hall (Transit Tunnel East / Undertransit Two)
    (0x66F6300E, 0x4BFCA639),
]

SANCTUARY_PAIRS = [
    # 01_cliff_entrance (Sanctuary Entrance / Hive Entrance)
    (0x47265C0B, 0xF3FA4121),
    # 0b_cliff_hall (Power Junction / Entrance Defense Hall)
    (0x31C07D16, 0xA70EA82B),
    # 0c_cliff_hall (Reactor Access / Hive Reactor Access)
    (0xB666B655, 0xB026CCEB),
    # 02_cliff_reactorcore (Reactor Core / Hive Reactor)
    (0x4A6CF295, 0xF8DBC03D),
    # 00_save_cliff_a (Save Station A / Hive Save Station 1)
    (0x65B801CE, 0xBB57222F),
    # 0d_cliff_hall (Minigyro Chamber / Unseen Way)
    (0x96F4CA1E, 0xD5FFF2AB),
    # 0m_cliff_hall (Transit Station / Hive Portal Chamber)
    (0xF7424EC3, 0x7B94B06B),
    # 00_pickup_cliff_a (Sanctuary Map Station / Hive Cache 1)
    (0x52BF89F3, 0x5F43767D),
    # 03_cliff_chicane (Hall of Combat Mastery / Culling Chamber)
    (0x5571E89E, 0x690B786A),
    # 04_cliff_contraption (Main Research / Staging Area)
    (0x914F1381, 0x5BBF334F),
    # 0q_cliff_hall (Agon Transport Access / Hazing Cliff)
    (0x750BBFEF, 0x37814F2A),
    # 0e_cliff_hall (Central Area Transport East / Central Hive East Transport)
    (0x1152015D, 0xC2D7966B),
    # 0l_cliff_hall (Central Area Transport West / Central Hive West Transport)
    (0x70E48580, 0x6CBCD4AB),
    # 06_cliff_spiderballguardian (Dynamo Works / Hive Dynamo Works)
    (0xC0113CE8, 0xF24EF6B4),
    # 08_cliff_ringtowers (Watch Station / Aerial Training Site)
    (0xA2406387, 0x40F7447E),
    # 0f_cliff_hall (Dynamo Access / Hive Dynamo Access)
    (0x42C85AD9, 0xFBAF3B2B),
    # 00_pickup_cliff_c (Dynamo Storage / Hive Cache 3)
    (0x1F7728F8, 0x072FCFBC),
    # 0n_cliff_hall (Sentinel's Path / Temple Security Access)
    (0xA4D81547, 0x42EC1D2B),
    # 0o_cliff_hall (Grand Abyss / Judgment Drop)
    (0x237EDE04, 0x55C479EB),
    # 07cliff_gyro (Main Gyro Chamber / Hive Gyro Chamber)
    (0x73342A54, 0xA3656292),
    # 09_cliff_temple (Sanctuary Temple / Hive Temple)
    (0x6C363CD4, 0xA92F00B3),
    # 05_cliff_engineroom (Vault / Vault Attack Portal)
    (0x52254361, 0x1DF13149),
    # 0g_cliff_hall (Temple Access / Hive Temple Access)
    (0xC56E919A, 0xEC875FEB),
    # 0i_cliff_hall (Checkpoint Station / Hive Gyro Access)
    (0x844A690C, 0x2735236B),
    # 00_save_cliff_b (Save Station B / Hive Save Station 2)
    (0xE32C7360, 0x22B5442E),
    # 00_cliff_generator_connect (Controller Access / Hive Controller Access)
    (0xE89BDC92, 0x26AC7F1E),
    # 0j_cliff_hall_bottom (Aerie Transport Station / Hive Ammo Station)
    (0xBAF94A13, 0xF1DF78DC),
    # 00_cliff_generator (Sanctuary Energy Controller / Hive Energy Controller)
    (0x0D032A6A, 0x2DECC318),
    # 10_cliff_rooftop (Aerie / Hive Summit)
    (0x5D3A0001, 0x9D804955),
]

GREAT_TEMPLE_PAIRS = [
    # 07_temple_mothlordsanctuary (Temple Sanctuary / Sanctum)
    (0xA2CB7E32, 0xD7C3B839),
    # 0h_temple_hall (Controller Transport / Sanctum Access)
    (0x11E50BC6, 0xCBAD889B),
    # 08_temple_sungenerator (Main Energy Controller / Sky Temple Energy Controller)
    (0x02A01334, 0x7B4AFA6F),
]



from open_prime_rando.echoes.specific_area_patches import rebalance_patches, required_fixes
from open_prime_rando.echoes.specific_area_patches.version_differences import EchoesVersion, patch_version_differences
from open_prime_rando.patcher_editor import PatcherEditor


def specific_patches(editor: PatcherEditor, area_patches: dict, legacy_compatibility: bool):
    if legacy_compatibility:
        required_fixes.torvus_temple(editor)
        required_fixes.command_center_door(editor)
    else:
        required_fixes.apply_all(editor)
        patch_version_differences(editor, EchoesVersion.NTSC_U)  # TODO: detect version
        if area_patches["rebalance_world"]:
            rebalance_patches.apply_all(editor)



import dataclasses
from collections.abc import Iterable

from ppc_asm.dol_file import DolFile
from retro_data_structures.game_check import Game


@dataclasses.dataclass(frozen=True)
class DolVersion:
    game: Game
    description: str
    build_string_address: int
    build_string: bytes
    sda2_base: int
    sda13_base: int


def find_version_for_dol(dol_file: DolFile, all_versions: Iterable[DolVersion]) -> DolVersion:
    dol_file.set_editable(False)
    with dol_file:
        for version in all_versions:
            build_string = dol_file.read(version.build_string_address, len(version.build_string))
            if (build_string[:6], build_string[6 + 16:]) == (version.build_string[:6], version.build_string[6 + 16:]):
                return version

    raise RuntimeError("Unsupported game version")



import json
import logging
from pathlib import Path

from retro_data_structures.asset_manager import FileProvider
from retro_data_structures.game_check import Game

from open_prime_rando.patcher_editor import PatcherEditor
from open_prime_rando.validator_with_default import DefaultValidatingDraft7Validator

LOG = logging.getLogger("p1r_patcher")


def _read_schema():
    with Path(__file__).parent.joinpath("prime_remastered", "schema.json").open() as f:
        return json.load(f)


def patch_paks(file_provider: FileProvider, output_path: Path, configuration: dict):
    LOG.info("Will patch files at %s", file_provider)
    editor = PatcherEditor(file_provider, Game.PRIME_REMASTER)

    LOG.info("Validating schema")
    DefaultValidatingDraft7Validator(_read_schema()).validate(configuration)

    # Save our changes
    editor.flush_modified_assets()

    editor.save_modifications(output_path)
    LOG.info("Finished.")



from random import Random

from open_prime_rando.echoes.small_randomizations.echo_locks import randomize_echo_locks
from open_prime_rando.echoes.small_randomizations.minigyro_chamber import randomize_minigyro_chamber
from open_prime_rando.echoes.small_randomizations.rubiks import randomize_rubiks_puzzles
from open_prime_rando.patcher_editor import PatcherEditor


def apply_small_randomizations(editor: PatcherEditor, configuration: dict):
    rng = Random(configuration["seed"])

    if configuration["echo_locks"]:
        randomize_echo_locks(editor, rng)

    if configuration["minigyro_chamber"]:
        randomize_minigyro_chamber(editor, rng)

    if configuration["rubiks"]:
        randomize_rubiks_puzzles(editor, rng)



import logging
from enum import Enum

from retro_data_structures.properties.echoes.archetypes.Transform import Transform
from retro_data_structures.properties.echoes.core.Vector import Vector
from retro_data_structures.properties.echoes.objects import Actor, Counter, Platform

from open_prime_rando.echoes.asset_ids import agon_wastes, great_temple, torvus_bog
from open_prime_rando.echoes.asset_ids.world import (
    AGON_WASTES_MLVL,
    GREAT_TEMPLE_MLVL,
    TORVUS_BOG_MLVL,
)
from open_prime_rando.patcher_editor import PatcherEditor

LOG = logging.getLogger("echoes_patcher")


class EchoesVersion(float, Enum):
    NTSC_U = 1.028
    PAL = 1.035
    NTSC_J = 1.036
    NEW_PLAY_CONTROL = 3.561
    TRILOGY_NTSC = 3.593
    TRILOGY_PAL = 3.629


def patch_version_differences(editor: PatcherEditor, version: EchoesVersion):
    """
    Patches version differences that affect gameplay or logic. Usually restores NTSC-U functionality.
    Note: only NTSC-U and PAL are officially supported. Version differences introduced in later versions
    do not strictly need to be patched.
    """
    transport_a_access(editor, version)
    path_of_eyes(editor, version)
    venomous_pond(editor, version)
    portal_terminal(editor, version)


def transport_a_access(editor: PatcherEditor, version: EchoesVersion):
    """
    Ensures the vulnerability for the blocker is correct.
    """
    if version == EchoesVersion.NTSC_U:
        return

    area = editor.get_area(GREAT_TEMPLE_MLVL, great_temple.TRANSPORT_A_ACCESS_MREA)

    with area.get_instance("blocker").edit_properties(Platform) as blocker:
        blocker.vulnerability.boost_ball.damage_multiplier = 100.0
        blocker.vulnerability.screw_attack.damage_multiplier = 100.0
        blocker.vulnerability.super_missle.damage_multiplier = 100.0


def path_of_eyes(editor: PatcherEditor, version: EchoesVersion):
    """
    Edits area collision to restore a gap you can boost through.
    """
    if version == EchoesVersion.NTSC_U:
        return

    LOG.warning(
        "Path of Eyes version differences have not been patched! "
        "Area collision edits are not yet implemented."
    )


def venomous_pond(editor: PatcherEditor, version: EchoesVersion):
    """
    Edits the collision of an Ing Bearerpod to restore a standable to reach the item.
    """
    if version == EchoesVersion.NTSC_U:
        return

    area = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.VENOMOUS_POND_MREA)

    with area.get_instance(0x10003E).edit_properties(Actor) as piggyplant:
        piggyplant.editor_properties.transform = Transform(
            position=Vector(17.589201, -148.844849, 37.929966),
            scale=Vector(1.5, 1.5, 1.5),
            rotation=piggyplant.editor_properties.transform.rotation
        )
        piggyplant.collision_model = 0x8E4170D5


def portal_terminal(editor: PatcherEditor, version: EchoesVersion):
    """
    In GC NTSC-J version a counter was added to check for each cork to be broken
    """
    if version < EchoesVersion.NTSC_J:
        return

    area = editor.get_area(AGON_WASTES_MLVL, agon_wastes.PORTAL_TERMINAL_MREA)

    counter = area.get_instance(0x12044E)

    # Remove counter increment on the 2 first corks to destroy
    relay_ids = [0x12033A, 0x120343]  # 0x120307 is the last cork to destroy
    for relay_id in relay_ids:
        relay = area.get_instance(relay_id)
        relay.remove_connections_from(counter)

    # Set the destroyed cork counter to expect only one cork to be destroyed
    with counter.edit_properties(Counter) as props:
        props.editor_properties.unknown = 1
        props.max_count = 1



# Generated by asset_id_files.py

CONTROLLER_TRANSPORT_MREA = 0x11E50BC6
MAIN_ENERGY_CONTROLLER_MREA = 0x02A01334
SANCTUM_MREA = 0xD7C3B839
SANCTUM_ACCESS_MREA = 0xCBAD889B
SKY_TEMPLE_ENERGY_CONTROLLER_MREA = 0x7B4AFA6F
TEMPLE_SANCTUARY_MREA = 0xA2CB7E32
TEMPLE_TRANSPORT_A_MREA = 0x185B40F0
TEMPLE_TRANSPORT_B_MREA = 0x8F01B104
TEMPLE_TRANSPORT_C_MREA = 0x9860CBB0
TRANSPORT_A_ACCESS_MREA = 0xE79AAFAE
TRANSPORT_B_ACCESS_MREA = 0xFEB7BD27
TRANSPORT_C_ACCESS_MREA = 0x1A4DDCC0

NAME_TO_ID_MREA = {
    "Controller Transport": 0x11E50BC6,
    "Main Energy Controller": 0x02A01334,
    "Sanctum": 0xD7C3B839,
    "Sanctum Access": 0xCBAD889B,
    "Sky Temple Energy Controller": 0x7B4AFA6F,
    "Temple Sanctuary": 0xA2CB7E32,
    "Temple Transport A": 0x185B40F0,
    "Temple Transport B": 0x8F01B104,
    "Temple Transport C": 0x9860CBB0,
    "Transport A Access": 0xE79AAFAE,
    "Transport B Access": 0xFEB7BD27,
    "Transport C Access": 0x1A4DDCC0,
}
# Generated by asset_id_files.py

CONTROLLER_TRANSPORT_MAPA = 0x3830F0FB
MAIN_ENERGY_CONTROLLER_MAPA = 0x2B75E809
SANCTUM_MAPA = 0xFE164304
SANCTUM_ACCESS_MAPA = 0xE27873A6
SKY_TEMPLE_ENERGY_CONTROLLER_MAPA = 0x529F0152
TEMPLE_SANCTUARY_MAPA = 0x8B1E850F
TEMPLE_TRANSPORT_A_MAPA = 0x318EBBCD
TEMPLE_TRANSPORT_B_MAPA = 0xA6D44A39
TEMPLE_TRANSPORT_C_MAPA = 0xB1B5308D
TRANSPORT_A_ACCESS_MAPA = 0xCE4F5493
TRANSPORT_B_ACCESS_MAPA = 0xD762461A
TRANSPORT_C_ACCESS_MAPA = 0x339827FD

NAME_TO_ID_MAPA = {
    "Controller Transport": 0x3830F0FB,
    "Main Energy Controller": 0x2B75E809,
    "Sanctum": 0xFE164304,
    "Sanctum Access": 0xE27873A6,
    "Sky Temple Energy Controller": 0x529F0152,
    "Temple Sanctuary": 0x8B1E850F,
    "Temple Transport A": 0x318EBBCD,
    "Temple Transport B": 0xA6D44A39,
    "Temple Transport C": 0xB1B5308D,
    "Transport A Access": 0xCE4F5493,
    "Transport B Access": 0xD762461A,
    "Transport C Access": 0x339827FD,
}
# Generated by asset_id_files.py

CONTROLLER_TRANSPORT_INTERNAL_ID = 0xFF8C5AD6
MAIN_ENERGY_CONTROLLER_INTERNAL_ID = 0x7E6249F4
SANCTUM_INTERNAL_ID = 0x1E265BC2
SANCTUM_ACCESS_INTERNAL_ID = 0x05484014
SKY_TEMPLE_ENERGY_CONTROLLER_INTERNAL_ID = 0x8D1B96EB
TEMPLE_SANCTUARY_INTERNAL_ID = 0xDC8B67D3
TEMPLE_TRANSPORT_A_INTERNAL_ID = 0x6B65085B
TEMPLE_TRANSPORT_B_INTERNAL_ID = 0x49BFB670
TEMPLE_TRANSPORT_C_INTERNAL_ID = 0x6546B2B9
TRANSPORT_A_ACCESS_INTERNAL_ID = 0x628F4FC2
TRANSPORT_B_ACCESS_INTERNAL_ID = 0x89F0943F
TRANSPORT_C_ACCESS_INTERNAL_ID = 0xFD0E6733

NAME_TO_ID_INTERNAL_ID = {
    "Controller Transport": 0xFF8C5AD6,
    "Main Energy Controller": 0x7E6249F4,
    "Sanctum": 0x1E265BC2,
    "Sanctum Access": 0x05484014,
    "Sky Temple Energy Controller": 0x8D1B96EB,
    "Temple Sanctuary": 0xDC8B67D3,
    "Temple Transport A": 0x6B65085B,
    "Temple Transport B": 0x49BFB670,
    "Temple Transport C": 0x6546B2B9,
    "Transport A Access": 0x628F4FC2,
    "Transport B Access": 0x89F0943F,
    "Transport C Access": 0xFD0E6733,
}

DOCK_NAMES = {
    "Controller Transport": {
        "NorthBottom": 0,
        "NorthTop": 1,
    },
    "Main Energy Controller": {
        "SouthLight": 0,
        "VirtualNorth": 1,
        "North1": 2,
    },
    "Sanctum": {
        "SouthDark": 0,
    },
    "Sanctum Access": {
        "DarkTop": 0,
        "Darkbottom": 1,
    },
    "Sky Temple Energy Controller": {
        "SouthLight": 0,
        "VirtualNorth": 1,
    },
    "Temple Sanctuary": {
        "North": 0,
        "West": 1,
        "South": 2,
        "East": 3,
    },
    "Temple Transport A": {
        "South": 0,
    },
    "Temple Transport B": {
        "West": 0,
    },
    "Temple Transport C": {
        "East": 0,
    },
    "Transport A Access": {
        "North": 0,
        "South": 1,
    },
    "Transport B Access": {
        "West": 0,
        "East": 1,
    },
    "Transport C Access": {
        "East": 0,
        "West": 1,
    },
}



import argparse
import importlib
import json
import logging
import logging.config
from pathlib import Path

from retro_data_structures.asset_manager import IsoFileProvider, PathFileProvider

_game_to_patcher = {
    "echoes": "open_prime_rando.echoes_patcher",
    "prime_remastered": "open_prime_rando.p1r_patcher",
}


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--game", required=True, choices=sorted(_game_to_patcher.keys()))
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument("--input-paks", type=Path,
                             help="Path to where the paks to randomize")
    input_group.add_argument("--input-iso", type=Path,
                             help="Path to a ISO to randomize")
    parser.add_argument("--output-paks", required=True, type=Path,
                        help="Path to where the modified paks will be written to.")
    parser.add_argument("--input-json", type=Path, required=True,
                        help="Path to the configuration json.")
    return parser


def setup_logging():
    handlers = {
        'default': {
            'level': 'DEBUG',
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
    }
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s',
            }
        },
        'handlers': handlers,
        'disable_existing_loggers': False,
        'loggers': {
            'default': {
                'level': 'DEBUG',
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': list(handlers.keys()),
        },
    })
    logging.info("Hello world.")


def main():
    setup_logging()
    parser = create_parser()
    args = parser.parse_args()
    print(args)

    with args.input_json.open() as f:
        configuration = json.load(f)

    if args.input_paks is not None:
        file_provider = PathFileProvider(args.input_paks)
    else:
        file_provider = IsoFileProvider(args.input_iso)

    patcher_module = importlib.import_module(_game_to_patcher[args.game])
    patcher_module.patch_paks(
        file_provider,
        args.output_paks,
        configuration,
    )



from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.strg import Strg
from retro_data_structures.properties.echoes.objects.WorldTeleporter import WorldTeleporter

from open_prime_rando.patcher_editor import PatcherEditor


def patch_elevator(editor: PatcherEditor, area: Area, elevator_id: int,
                   target_mlvl: int, target_mrea: int, target_strg: int, target_name: str):

    elevator = area.get_instance(elevator_id)
    with elevator.edit_properties(WorldTeleporter) as props:
        props.world = target_mlvl
        props.area = target_mrea

    if target_strg is not None:
        strg = editor.get_file(target_strg, Strg)
        strg.set_string(1, f"Access to &push;&main-color=#FF3333;{target_name} &pop;granted. "
                           f"Step into the hologram to activate elevator.")



import dataclasses

from retro_data_structures.asset_manager import NameOrAssetId
from retro_data_structures.base_resource import AssetId
from retro_data_structures.enums.echoes import Message, State, VisorFlags
from retro_data_structures.formats.mapa import Mapa
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.scan import Scan
from retro_data_structures.formats.script_object import ScriptInstance
from retro_data_structures.formats.strg import Strg
from retro_data_structures.properties.base_property import BaseObjectType
from retro_data_structures.properties.echoes.archetypes.ActorParameters import ActorParameters
from retro_data_structures.properties.echoes.archetypes.DamageVulnerability import DamageVulnerability
from retro_data_structures.properties.echoes.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.echoes.archetypes.HealthInfo import HealthInfo
from retro_data_structures.properties.echoes.archetypes.ScannableParameters import ScannableParameters
from retro_data_structures.properties.echoes.archetypes.SurroundPan import SurroundPan
from retro_data_structures.properties.echoes.archetypes.Transform import Transform
from retro_data_structures.properties.echoes.archetypes.VisorParameters import VisorParameters
from retro_data_structures.properties.echoes.archetypes.WeaponVulnerability import WeaponVulnerability
from retro_data_structures.properties.echoes.core.Color import Color
from retro_data_structures.properties.echoes.core.Spline import Spline
from retro_data_structures.properties.echoes.core.Vector import Vector
from retro_data_structures.properties.echoes.objects.Actor import Actor
from retro_data_structures.properties.echoes.objects.CameraShaker import CameraShaker
from retro_data_structures.properties.echoes.objects.Counter import Counter
from retro_data_structures.properties.echoes.objects.DamageableTrigger import DamageableTrigger
from retro_data_structures.properties.echoes.objects.DamageableTriggerOrientated import DamageableTriggerOrientated
from retro_data_structures.properties.echoes.objects.Dock import Dock
from retro_data_structures.properties.echoes.objects.Door import Door
from retro_data_structures.properties.echoes.objects.Effect import Effect
from retro_data_structures.properties.echoes.objects.HUDHint import HUDHint
from retro_data_structures.properties.echoes.objects.MemoryRelay import MemoryRelay
from retro_data_structures.properties.echoes.objects.PlayerController import PlayerController
from retro_data_structures.properties.echoes.objects.Relay import Relay
from retro_data_structures.properties.echoes.objects.ScannableObjectInfo import ScannableObjectInfo
from retro_data_structures.properties.echoes.objects.Sound import Sound
from retro_data_structures.properties.echoes.objects.StreamedAudio import StreamedAudio
from retro_data_structures.properties.echoes.objects.Timer import Timer

from open_prime_rando.echoes.asset_ids import world
from open_prime_rando.echoes.dock_lock_rando.map_icons import DoorMapIcon
from open_prime_rando.echoes.vulnerabilities import resist_all_vuln
from open_prime_rando.patcher_editor import PatcherEditor


@dataclasses.dataclass(kw_only=True)
class DoorType:
    name: str

    vulnerability: DamageVulnerability

    shell_model: NameOrAssetId = 0x6B78FD92  # normal door model
    shell_color: Color

    scan_text: tuple[str, ...] | None = None

    map_icon: DoorMapIcon

    patched_scan: AssetId | None = None

    def get_files(self, editor: PatcherEditor, world_name: str, area_name: str) -> tuple[Area, Mapa]:
        world_file = world.load_dedicated_file(world_name)
        area = editor.get_area(world.NAME_TO_ID_MLVL[world_name], world_file.NAME_TO_ID_MREA[area_name])
        mapa = editor.get_file(world_file.NAME_TO_ID_MAPA[area_name], type_hint=Mapa)
        return area, mapa

    def get_dock_index(self, world_name: str, area_name: str, dock_name: str) -> int:
        world_file = world.load_dedicated_file(world_name)
        return world_file.DOCK_NAMES[area_name][dock_name]

    def get_area(self, editor: PatcherEditor, world_name: str, area_name: str) -> Area:
        return self.get_files(editor, world_name, area_name)[0]

    def get_door_from_dock_index(self, area: Area, dock_index: int) -> ScriptInstance:
        dock = next(
            inst for inst in area.all_instances
            if (
                    inst.type == Dock and
                    inst.get_properties_as(Dock).dock_number == dock_index
            )
        )
        for instance in area.all_instances:
            if instance.type != Door:
                continue
            for connection in instance.connections:
                if dock.id_matches(connection.target):
                    return instance
        raise KeyError(f"no door found with a connection to {dock} in {area.name}")

    def patch_map_icon(self, mapa: Mapa, door: ScriptInstance):
        for obj in mapa.raw.mappable_objects:
            if door.id_matches(obj.editor_id):
                obj.type = self.map_icon.value
                return

    @staticmethod
    def get_scan_templates(editor: PatcherEditor) -> tuple[Scan, Strg]:
        # Uncategorized/There is a Blast Shield on the door blocking acces_0.SCAN
        scan = editor.get_parsed_asset(0x36DE1342, type_hint=Scan)
        # Strings/Uncategorized/There is a Blast Shield on the door blocking acces_0_0.STRG
        strg = editor.get_parsed_asset(0x49DF4448, type_hint=Strg)
        return scan, strg

    def get_patched_scan(self, editor: PatcherEditor, world_name: str, area_name: str) -> AssetId:
        if self.patched_scan is None or not editor.does_asset_exists(self.patched_scan):
            scan, strg = DoorType.get_scan_templates(editor)
            for i, text in enumerate(self.scan_text):
                strg.set_string(i, text)

            strg_id = editor.add_or_replace_custom_asset(f"custom_door_{self.name}.STRG", strg)

            with scan.scannable_object_info.edit_properties(ScannableObjectInfo) as scan_info:
                scan_info.string = strg_id

            scan_id = editor.add_or_replace_custom_asset(f"custom_door_{self.name}.SCAN", scan)
            self.patched_scan = scan_id

        return self.patched_scan

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        area, mapa = self.get_files(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))
        self.patch_map_icon(mapa, door)

        shell_model = editor._resolve_asset_id(self.shell_model)

        with door.edit_properties(Door) as door_props:
            door_props.shell_model = shell_model
            door_props.shell_color = self.shell_color
            if self.scan_text is not None:
                door_props.alt_scannable.scannable_info0 = self.get_patched_scan(editor, world_name, area_name)
            else:
                door_props.alt_scannable.scannable_info0 = 0xFFFFFFFF


@dataclasses.dataclass(kw_only=True)
class NormalDoorType(DoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        with door.edit_properties(Door) as door_props:
            door_props.vulnerability = self.vulnerability


@dataclasses.dataclass
class BlastShieldActors:
    door: ScriptInstance
    sound: ScriptInstance | None
    streamed: ScriptInstance | None
    lock: ScriptInstance
    relay: ScriptInstance
    gibs: ScriptInstance | None


@dataclasses.dataclass(kw_only=True)
class BlastShieldDoorType(DoorType):
    shield_model: NameOrAssetId
    shield_collision_box: Vector = dataclasses.field(default_factory=lambda: Vector(0.35, 5.0, 4.0))
    shield_collision_offset: Vector = dataclasses.field(default_factory=lambda: Vector(-2 / 3, 0, 2.0))

    def find_attached_instance(
            self, area: Area, source: ScriptInstance, state: State, message: Message,
            target_type: type[BaseObjectType], target_name: str | None = None
    ) -> ScriptInstance:
        for connection in source.connections:
            if connection.state == state and connection.message == message:
                target = area.get_instance(connection.target)
                if target.type == target_type and (target_name is None or target.name == target_name):
                    return target
        name = f"{target_type} named {target_name}" if target_name is not None else str(target_type)
        raise TypeError(f"No {name} connected to {source}")

    def get_spline(self) -> Spline:
        return Spline.from_bytes(
            b'\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x02'
            b'\x02A \x00\x00?\x80\x00\x00\x02\x02\x01\x00\x00\x00\x00?\x80\x00\x00'
        )

    def create_trigger(self,
                       name: str,
                       door_xfm: Transform,
                       health: float,
                       visor_flags: VisorFlags = VisorFlags.Combat | VisorFlags.Dark,
                       active: bool = True,
                       seeker_lock_on: bool = True,
                       orbitable: bool = False,
                       ):
        pos = Vector(
            door_xfm.position.x,
            door_xfm.position.y,
            door_xfm.position.z + 1.8
        )

        return DamageableTriggerOrientated(
            editor_properties=EditorProperties(
                name=name,
                transform=Transform(
                    position=pos,
                    rotation=door_xfm.rotation,
                    scale=Vector(4.0, 4.0, 1.5)
                ),
                active=active
            ),
            health=HealthInfo(health=health),
            vulnerability=self.vulnerability,
            enable_seeker_lock_on=seeker_lock_on,
            orbitable=orbitable,
            visor=VisorParameters(
                visor_flags=visor_flags
            )
        )

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        """
        blast shield connections:
            DEAD -> lock cleared MemoryRelay, ACTV
            DEAD -> lock breaking SoundEffect, PLAY
            DEAD -> lock explosion gibs, ACTV
            DEAD -> jingle streamedaudio, PLAY

        door connections:
            OPEN -> lock cleared MemoryRelay, ACTV

        memory relay connections:
            ACTV -> door, RSET
            ACTV -> blast shield, DCTV
        """
        super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        with door.edit_properties(Door) as _door:
            door_transform = _door.editor_properties.transform
            _door.vulnerability = resist_all_vuln

        default = area.get_layer("Default")

        shield_model = self.shield_model
        if isinstance(shield_model, str):
            shield_model = f"custom_door_lock_{shield_model}.CMDL"
        model = editor._resolve_asset_id(shield_model)

        lock = default.add_instance_with(Actor(
            editor_properties=EditorProperties(
                name=f"{self.name} Blast Shield Lock",
                transform=door_transform
            ),
            collision_box=self.shield_collision_box,
            collision_offset=self.shield_collision_offset,
            vulnerability=self.vulnerability,
            model=model,
            actor_information=ActorParameters(
                scannable=ScannableParameters(
                    scannable_info0=self.get_patched_scan(editor, world_name, area_name)
                )
            )
        ))

        relay = default.add_memory_relay("Lock Cleared")
        with relay.edit_properties(MemoryRelay) as _relay:
            _relay.editor_properties.transform = door_transform
            _relay.editor_properties.active = False

        lock.add_connection(State.Dead, Message.Activate, relay)

        relay.add_connection(State.Active, Message.Deactivate, lock)
        relay.add_connection(State.Active, Message.Reset, door)

        door.add_connection(State.Open, Message.Activate, relay)

        gibs = None
        sound = None
        streamed = None

        if not low_memory:
            particle = 0xCDCBDF04

            gibs = default.add_instance_with(Effect(
                editor_properties=EditorProperties(
                    transform=door_transform,
                    active=False
                ),
                particle_effect=particle,
                restart_on_activate=True,
                motion_control_spline=self.get_spline(),
            ))

            sound = default.add_instance_with(Sound(
                editor_properties=EditorProperties(
                    name="Metal Door Lock Breaking",
                    transform=door_transform
                ),
                sound=948,
                max_audible_distance=150.0,
                surround_pan=SurroundPan(surround_pan=1.0)
            ))

            streamed = default.add_instance_with(StreamedAudio(
                editor_properties=EditorProperties(
                    name="StreamedAudio - Event Jingle",
                    transform=door_transform
                ),
                song_file="/audio/evt_x_event_00.dsp",
                fade_in_time=0.01,
                volume=65,
                software_channel=1,
                unknown=False
            ))

            lock.add_connection(State.Dead, Message.Play, sound)
            lock.add_connection(State.Dead, Message.Play, streamed)
            lock.add_connection(State.Dead, Message.Activate, gibs)

        return BlastShieldActors(door, sound, streamed, lock, relay, gibs)


@dataclasses.dataclass(kw_only=True)
class VanillaBlastShieldDoorType(BlastShieldDoorType):
    def remove_blast_shield(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str):
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        relay = self.find_attached_instance(area, door, State.Open, Message.Activate, MemoryRelay)
        lock = self.find_attached_instance(area, relay, State.Active, Message.Deactivate, Actor)

        for connection in lock.connections:
            area.remove_instance(connection.target)
        area.remove_instance(lock)


@dataclasses.dataclass(kw_only=True)
class SeekerBlastShieldDoorType(VanillaBlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        actors = super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        default = area.get_layer("Default")

        # immune instead of reflect so that it explodes on the lock
        missile_immune = dataclasses.replace(resist_all_vuln, missile=WeaponVulnerability(damage_multiplier=0.0))
        with actors.lock.edit_properties(Actor) as lock:
            lock.vulnerability = missile_immune
        with actors.door.edit_properties(Door) as door:
            door.vulnerability = missile_immune

        door_transform = actors.door.get_properties_as(Door).editor_properties.transform

        # TODO: try using a sequencetimer to squeeze one fewer instance out of this
        timer = default.add_instance_with(Timer(
            editor_properties=EditorProperties(
                name="Button Control",
                transform=door_transform
            ),
            time=0.75
        ))

        timer_reset = default.add_instance_with(Timer(
            editor_properties=EditorProperties(
                name="Button Reset",
                transform=door_transform
            ),
            time=0.01
        ))

        # create 5 triggers so that you can have 5 lock-ons
        # TODO: only use 2 triggers in low memory mode and just accept it being more jank?
        # 30.01 health because splash damage is inconsistent. missiles do 30 damage
        # so this guarantees you need at least 2 missiles at once to break it
        triggers = [
            default.add_instance_with(self.create_trigger(
                f"Bridge Button {i}",
                door_transform, 30.01)
            ) for i in range(5)
        ]
        main_trigger = triggers[0]
        # TODO: repurpose one of the big triggers for this to save an instance
        mini_trigger = default.add_instance_with(self.create_trigger(
            "Bridge Button Mini",
            door_transform,
            1.0,
            seeker_lock_on=False
        ))

        # start a timer when the tiny trigger dies. stop it if the main trigger dies
        mini_trigger.add_connection(State.Dead, Message.ResetAndStart, timer)
        main_trigger.add_connection(State.Dead, Message.Stop, timer)
        for trigger in triggers:
            timer.add_connection(State.Zero, Message.Deactivate, trigger)

        # if the timer counts all the way down, reset the triggers
        timer.add_connection(State.Zero, Message.ResetAndStart, timer_reset)
        timer_reset.add_connection(State.Zero, Message.Activate, mini_trigger)
        for trigger in triggers:
            timer_reset.add_connection(State.Zero, Message.Activate, trigger)

        # move the lock's connections to the trigger, since it's the thing that dies now
        for connection in actors.lock.connections:
            actors.lock.remove_connection(connection)
            main_trigger.add_connection(
                connection.state,
                connection.message,
                connection.target
            )

        for trigger in triggers:
            actors.relay.add_connection(State.Active, Message.Deactivate, trigger)
        actors.relay.add_connection(State.Active, Message.Deactivate, mini_trigger)
        actors.relay.add_connection(State.Active, Message.Deactivate, timer)
        actors.relay.add_connection(State.Active, Message.Deactivate, timer_reset)

    def remove_blast_shield(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str):
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        memory_relay = self.find_attached_instance(area, door, State.Open, Message.Activate, MemoryRelay)
        trigger = self.find_attached_instance(area, memory_relay, State.Active, Message.Deactivate, DamageableTrigger)
        shaker = self.find_attached_instance(area, trigger, State.Dead, Message.Action, CameraShaker)
        counter = self.find_attached_instance(area, trigger, State.Dead, Message.Increment, Counter)
        complete_relay = self.find_attached_instance(area, counter, State.MaxReached, Message.SetToZero, Relay)

        area.remove_instance(shaker)
        for connection in complete_relay.connections:
            area.remove_instance(connection.target)
        area.remove_instance(complete_relay)
        area.remove_instance(counter)


@dataclasses.dataclass(kw_only=True)
class PlanetaryEnergyDoorType(DoorType):
    planetary_energy_item_id: int

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class GrappleDoorType(BlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass
class VisorBlastShieldActors(BlastShieldActors):
    trigger: ScriptInstance
    visor_detector: ScriptInstance


@dataclasses.dataclass(kw_only=True)
class VisorDoorType(BlastShieldDoorType):
    visor_flags: VisorFlags
    player_controller_proxy: int

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        actors = super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        default = area.get_layer("Default")

        with actors.door.edit_properties(Door) as door:
            door.vulnerability = resist_all_vuln
            door_xfm = door.editor_properties.transform

        with actors.lock.edit_properties(Actor) as lock:
            lock.vulnerability = resist_all_vuln

        trigger = default.add_instance_with(self.create_trigger(
            "Door Button", door_xfm, 1.0, self.visor_flags,
            active=False, seeker_lock_on=False, orbitable=True
        ))

        visor_detector = default.add_instance_with(PlayerController(
            editor_properties=EditorProperties(
                name="Detect Visor",
                transform=door_xfm
            ),
            unknown_0xe71de331=1,
            proxy_type=self.player_controller_proxy
        ))

        visor_detector.add_connection(State.Entered, Message.Activate, trigger)
        visor_detector.add_connection(State.Exited, Message.Deactivate, trigger)

        for connection in actors.lock.connections:
            actors.lock.remove_connection(connection)
            trigger.add_connection(
                connection.state,
                connection.message,
                connection.target
            )

        actors.relay.add_connection(State.Active, Message.Deactivate, trigger)
        actors.relay.add_connection(State.Active, Message.Deactivate, visor_detector)

        return VisorBlastShieldActors(
            actors.door, actors.sound, actors.streamed,
            actors.lock, actors.relay, actors.gibs,
            trigger, visor_detector
        )


@dataclasses.dataclass(kw_only=True)
class DarkVisorDoorType(VisorDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        actors = super().patch_door(editor, world_name, area_name, dock_name, low_memory)

        with actors.lock.edit_properties(Actor) as lock:
            # TODO: update the template
            # this property makes the actor appear red in dark visor
            lock.actor_information.unknown_0xcd4c81a1 = True


@dataclasses.dataclass(kw_only=True)
class EchoVisorDoorType(VisorDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        actors = super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        default = area.get_layer("Default")

        door_xfm = actors.door.get_properties_as(Door).editor_properties.transform

        center_pos = Vector(
            door_xfm.position.x,
            door_xfm.position.y,
            door_xfm.position.z + 1.8
        )

        hud_hint = default.add_instance_with(HUDHint(
            editor_properties=EditorProperties(
                name="Echo Target Icon",
                transform=Transform(
                    center_pos,
                    door_xfm.rotation,
                ),
                active=True
            ),
            hud_texture=0x36B1CB06,
            unknown_0x6078a651=24.0,
            unknown_0xf00bb6bb=128.0,
            animation_time=0.5,
            animation_frames=4,
            unknown_0xd993f97b=2
        ))

        with actors.lock.edit_properties(Actor) as lock:
            lock.echo_information.is_echo_emitter = True

        actors.relay.add_connection(State.Active, Message.Deactivate, hud_hint)

        if not low_memory:
            beacon_loop = default.add_instance_with(Sound(
                editor_properties=EditorProperties(
                    name="Echo Beacon Sound Loop",
                    transform=Transform(
                        position=center_pos
                    ),
                ),
                sound=1003,
                max_audible_distance=75.0,
                min_volume=0,
                max_volume=60,
                surround_pan=SurroundPan(
                    pan=0.0,
                    surround_pan=1.0
                ),
                loop=True,
                play_always=True,
                echo_visor_max_volume=127
            ))

            disrupted = default.add_instance_with(Sound(
                editor_properties=EditorProperties(
                    name="Echo Particle Disrupted",
                    transform=Transform(
                        position=center_pos
                    ),
                ),
                sound=1004,
                max_audible_distance=75.0,
                min_volume=20,
                max_volume=120,
                surround_pan=SurroundPan(
                    pan=0.0,
                    surround_pan=1.0
                ),
            ))

            timer = default.add_instance_with(Timer(
                editor_properties=EditorProperties(
                    name="Open Echo Door",
                    transform=door_xfm,
                ),
                time=1.0
            ))

            for connection in actors.trigger.connections:
                actors.trigger.remove_connection(connection)
                timer.add_connection(
                    State.Zero,
                    connection.message,
                    connection.target
                )

            actors.trigger.add_connection(State.Dead, Message.ResetAndStart, timer)
            actors.trigger.add_connection(State.Dead, Message.InternalMessage00, hud_hint)
            actors.trigger.add_connection(State.Dead, Message.Stop, beacon_loop)
            actors.trigger.add_connection(State.Dead, Message.Play, disrupted)

            actors.visor_detector.add_connection(State.Entered, Message.Play, beacon_loop)
            actors.visor_detector.add_connection(State.Exited, Message.Stop, beacon_loop)

            actors.relay.add_connection(State.Active, Message.Deactivate, beacon_loop)
            actors.relay.add_connection(State.Active, Message.Deactivate, timer)
            actors.relay.add_connection(State.Active, Message.Deactivate, disrupted)


@dataclasses.dataclass(kw_only=True)
class ScanVisorDoorType(BlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class TranslatorDoorType(ScanVisorDoorType):
    translator_item_id: int

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()



from retro_data_structures.base_resource import RawResource

from open_prime_rando.echoes.custom_assets import custom_asset_path
from open_prime_rando.echoes.suit_cosmetics.asset_map import SUIT_ASSETS
from open_prime_rando.patcher_editor import PatcherEditor


def apply_custom_suits(editor: PatcherEditor, configuration: dict):
    for suit, assets in SUIT_ASSETS.items():
        skin = configuration[suit]
        if skin == "player1":
            continue
        custom_suit_assets = custom_asset_path().joinpath("suits", skin, suit)
        for asset_id, filename in assets.items():
            asset = custom_suit_assets.joinpath(filename)
            if not asset.exists():
                continue # some skins leave a few assets vanilla

            res = RawResource(
                type="TXTR",
                data=asset.read_bytes(),
            )
            editor.replace_asset(asset_id, res)




import json
import logging
from collections.abc import Callable
from pathlib import Path

from retro_data_structures.asset_manager import FileProvider
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.strg import Strg
from retro_data_structures.game_check import Game

from open_prime_rando import dynamic_schema
from open_prime_rando.echoes import asset_ids, custom_assets, dock_lock_rando, specific_area_patches
from open_prime_rando.echoes.elevators import auto_enabled_elevator_patches
from open_prime_rando.echoes.elevators.elevator_rando import patch_elevator
from open_prime_rando.echoes.inverted import apply_inverted
from open_prime_rando.echoes.small_randomizations import apply_small_randomizations
from open_prime_rando.echoes.suit_cosmetics import apply_custom_suits
from open_prime_rando.patcher_editor import PatcherEditor
from open_prime_rando.validator_with_default import DefaultValidatingDraft7Validator

LOG = logging.getLogger("echoes_patcher")


def _read_schema():
    with Path(__file__).parent.joinpath("echoes", "schema.json").open() as f:
        return json.load(f)


def apply_area_modifications(editor: PatcherEditor, configuration: dict[str, dict],
                             status_update: Callable[[str, float], None]):
    num_areas = sum(len(world_config["areas"]) for world_config in configuration.values())
    areas_processed = 0.0

    for world_name, world_config in configuration.items():
        world_meta = asset_ids.world.load_dedicated_file(world_name)
        mlvl = editor.get_mlvl(asset_ids.world.NAME_TO_ID_MLVL[world_name])

        mrea_to_name: dict[int, str] = {
            mrea: name
            for name, mrea in world_meta.NAME_TO_ID_MREA.items()
        }

        areas_by_name: dict[str, Area] = {
            mrea_to_name[area.mrea_asset_id]: area
            for area in mlvl.areas
        }

        for i, (area_name, area) in enumerate(areas_by_name.items()):
            if area_name not in world_config["areas"]:
                continue

            status_update(f"Processing {world_name} - {area_name}...", areas_processed/num_areas)
            areas_processed += 1

            area_config = world_config["areas"][area_name]
            low_memory = area_config["low_memory_mode"]

            for dock_name, dock_config in area_config["docks"].items():
                dock_number = world_meta.DOCK_NAMES[area_name][dock_name]

                if "new_door_type" in dock_config:
                    dock_lock_rando.apply_door_rando(
                        editor,
                        world_name,
                        area_name,
                        dock_name,
                        dock_config["new_door_type"],
                        dock_config.get("old_door_type"),
                        low_memory
                    )

                if "connect_to" in dock_config:
                    dock_target = dock_config["connect_to"]
                    LOG.debug("Connecting dock %s of %s - %s to %s - %s",
                              dock_name, world_name, area_name, dock_target["area"], dock_target["dock"])
                    area.connect_dock_to(dock_number, areas_by_name[dock_target["area"]],
                                         world_meta.DOCK_NAMES[dock_target["area"]][dock_target["dock"]])

            for layer_name, layer_state in area_config["layers"].items():
                LOG.debug("Setting layer %s of %s - %s to %s", layer_name, world_name, area_name, str(layer_state))
                area.get_layer(layer_name).active = layer_state

            for elevator in area_config["elevators"]:
                patch_elevator(
                    editor,
                    area,
                    elevator["instance_id"],
                    elevator["target_assets"]["world_asset_id"],
                    elevator["target_assets"]["area_asset_id"],
                    elevator["target_strg"],
                    elevator["target_name"]
                )

            if area_config["new_name"] is not None:
                old_strg = area._raw.area_name_id
                strg = editor.get_parsed_asset(old_strg, type_hint=Strg)
                strg.set_string(0, area_config["new_name"])
                new_strg = editor.add_new_asset(f"custom_name_for_{area.internal_name}.STRG", strg)
                area._raw.area_name_id = new_strg

            area.update_all_dependencies(only_modified=True)


def apply_corrupted_memory_card_change(editor: PatcherEditor):
    # STRG_MemoryCard_0
    table = editor.get_file(0x88E242D6, Strg)

    name_to_index = {
        table.raw.name_table.name_array[entry.offset].string: entry.index
        for entry in table.raw.name_table.name_entries
    }

    table.set_string(
        name_to_index["CorruptedFile"],
        """The save file was created using a different
Randomizer ISO and must be deleted."""
    )
    table.set_string(
        name_to_index["ChoiceDeleteCorruptedFile"],
        "Delete Incompatible File"
    )


def patch_paks(file_provider: FileProvider,
               output_path: Path,
               configuration: dict,
               status_update: Callable[[str, float], None] = lambda s, _: LOG.info(s)):
    status_update(f"Will patch files at {file_provider}", 0)
    output_path.joinpath("files").mkdir(parents=True, exist_ok=True)
    output_path.joinpath("files", "opr_patcher_data.json").write_text(json.dumps(configuration))

    editor = PatcherEditor(file_provider, Game.ECHOES)

    status_update("Preparing schema", 0)
    schema = dynamic_schema.expand_schema(_read_schema(), editor)

    status_update("Validating schema", 0)
    DefaultValidatingDraft7Validator(schema).validate(configuration)

    legacy_compatibility: bool = configuration["legacy_compatibility"]

    status_update("Applying small patches", 0)
    if not legacy_compatibility:
        custom_assets.create_custom_assets(editor, include_premade=True)

    dock_lock_rando.add_custom_models(editor)
    if configuration["auto_enabled_elevators"]:
        auto_enabled_elevator_patches.apply_auto_enabled_elevators_patch(editor)
    specific_area_patches.specific_patches(editor, configuration["area_patches"], legacy_compatibility)
    apply_small_randomizations(editor, configuration["small_randomizations"])
    apply_corrupted_memory_card_change(editor)

    status_update("Modifying areas", 0)
    apply_area_modifications(editor, configuration["worlds"], status_update)

    if configuration["inverted"]:
        apply_inverted(editor)

    apply_custom_suits(editor, configuration["cosmetics"]["suits"])

    # Save our changes
    editor.flush_modified_assets()

    editor.save_modifications(output_path)
    status_update("Finished", 1.0)



from . import agon_wastes, great_temple, sanctuary_fortress, temple_grounds, torvus_bog, world

__all__ = [
    "great_temple",
    "sanctuary_fortress",
    "torvus_bog",
    "agon_wastes",
    "temple_grounds",
    "world",
]



from unittest.mock import MagicMock, call

from open_prime_rando.echoes import custom_assets


def test_create_split_ammo():
    manager = MagicMock()
    asset = manager.get_parsed_asset.return_value
    find_paks = manager.find_paks.return_value

    custom_assets._create_split_ammo(manager)

    # Assert
    manager.add_new_asset.assert_has_calls([
        call('dark_ammo_cmdl', asset, find_paks),
        call('dark_ammo_ancs', asset, find_paks),
        call('light_ammo_cmdl', asset, find_paks),
        call('light_ammo_ancs', asset, find_paks)
    ])



from retro_data_structures.formats.mrea import Area

CUSTOM_AREA_NAMES = {
    0xF3EE585F: "Portal Chamber (Light)",
    0xAE1E1339: "Portal Chamber (Dark)",
}


def get_name_for_area(area: Area) -> str:
    return CUSTOM_AREA_NAMES.get(area.mrea_asset_id, area.name)



import dataclasses


@dataclasses.dataclass(frozen=True)
class BeamAmmoConfiguration:
    item_index: int = dataclasses.field(metadata={"min": 0, "max": 108})
    ammo_a: int = dataclasses.field(metadata={"min": -1, "max": 108})
    ammo_b: int = dataclasses.field(metadata={"min": -1, "max": 108})
    uncharged_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 1})
    charged_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 5})
    combo_missile_cost: int = dataclasses.field(metadata={"min": 1, "max": 250, "if_different": 5})
    combo_ammo_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 30})

    @classmethod
    def from_json(cls, data: dict) -> "BeamAmmoConfiguration":
        return cls(
            item_index=data["item_index"],
            ammo_a=data["ammo_a"],
            ammo_b=data["ammo_b"],
            uncharged_cost=data["uncharged_cost"],
            charged_cost=data["charged_cost"],
            combo_missile_cost=data["combo_missile_cost"],
            combo_ammo_cost=data["combo_ammo_cost"],
        )
